import argparse  # noqa
import atexit  # noqa
import codecs  # noqa
import copy  # noqa
import errno  # noqa
import fnmatch  # noqa
import hashlib  # noqa
import os  # noqa
import shutil  # noqa
import signal  # noqa
import sys  # noqa
import threading  # noqa
import traceback  # noqa
from contextlib import contextmanager  # noqa
from datetime import datetime, timedelta  # noqa
from typing import (  # noqa
    TYPE_CHECKING,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

import sqlalchemy  # noqa
import yaml  # noqa
from loguru import logger  # noqa
from sqlalchemy.engine import Engine
from sqlalchemy.exc import OperationalError  # noqa
from sqlalchemy.ext.declarative import declarative_base  # noqa
from sqlalchemy.orm import sessionmaker  # noqa

# These need to be declared before we start importing from other flexget modules, since they might import them
from flexget.config_schema import ConfigError
from flexget.utils.sqlalchemy_utils import ContextSession  # noqa
from flexget.utils.tools import get_current_flexget_version, io_encoding, pid_exists  # noqa

Base = declarative_base()
Session: Type[ContextSession] = sessionmaker(class_=ContextSession)

import flexget.log  # noqa
from flexget import config_schema, db_schema, plugin  # noqa
from flexget.event import fire_event  # noqa
from flexget.ipc import IPCClient, IPCServer  # noqa
from flexget.options import (  # noqa
    CoreArgumentParser,
    ParserError,
    get_parser,
    manager_parser,
    unicode_argv,
)
from flexget.task import Task  # noqa
from flexget.task_queue import TaskQueue  # noqa
from flexget.terminal import console, get_console_output  # noqa

if TYPE_CHECKING:
    from flexget.tray_icon import TrayIcon
    from flexget.utils.simple_persistence import SimplePersistence

logger = logger.bind(name='manager')

manager: Optional['Manager'] = None
DB_CLEANUP_INTERVAL = timedelta(days=7)


class Manager:
    """Manager class for FlexGet

    Fires events:

    * manager.initialize

      The first time the manager is initialized, before config is loaded

    * manager.before_config_load

      Before the config file is loaded from disk

    * manager.before_config_validate

      When updating the config, before the validator is run on it

    * manager.config_updated

      After a configuration file has been loaded or changed (and validated) this event is fired

    * manager.startup

      After manager has been initialized. This is when application becomes ready to use,
      however no database lock is present, so the database must not be modified on this event.

    * manager.lock_acquired

      The manager does not always require a lock on startup, if one is requested,
      this event will run when it has been acquired successfully

    * manager.upgrade

      If any plugins have declared a newer schema version than exists in the database,
      this event will be fired to allow plugins to upgrade their tables

    * manager.shutdown_requested

      When shutdown has been requested. Any plugins which might add to
      execution queue should stop when this is fired.

    * manager.shutdown

      When the manager is exiting

    * manager.execute.completed

      If execution in current process was completed

    * manager.daemon.started
    * manager.daemon.completed
    * manager.db_cleanup
    """

    unit_test = False
    options: argparse.Namespace

    def __init__(self, args: Optional[List[str]]) -> None:
        """
        :param args: CLI args
        """
        global manager
        if not self.unit_test:
            assert not manager, 'Only one instance of Manager should be created at a time!'
        elif manager:
            logger.info('last manager was not torn down correctly')

        if args is None:
            # Decode all arguments to unicode before parsing
            args = unicode_argv()[1:]
        self.args = args
        self.autoreload_config = False
        self.config_file_hash: Optional[str] = None
        self.config_base: str = ''
        self.config_name: str = ''
        self.config_path: str = ''
        self.log_filename: str = ''
        self.db_filename: str = ''
        self.engine: Optional[Engine] = None
        self.lockfile: str = ''
        self.database_uri: str = ''
        self.db_upgraded = False
        self._has_lock = False
        self.is_daemon = False
        self.ipc_server: IPCServer
        self.task_queue: TaskQueue
        self.persist: 'SimplePersistence'
        self.initialized = False

        self.config: Dict = {}

        self.options = self._init_options(self.args)
        try:
            self._init_config(create=False)
        except Exception:
            flexget.log.start(level=self.options.loglevel, to_file=False)
            raise

        manager = self

        logger.debug('sys.defaultencoding: {}', sys.getdefaultencoding())
        logger.debug('sys.getfilesystemencoding: {}', sys.getfilesystemencoding())
        logger.debug('flexget detected io encoding: {}', io_encoding)
        logger.debug('os.path.supports_unicode_filenames: {}', os.path.supports_unicode_filenames)
        if (
            codecs.lookup(sys.getfilesystemencoding()).name == 'ascii'
            and not os.path.supports_unicode_filenames
        ):
            logger.warning(
                'Your locale declares ascii as the filesystem encoding. Any plugins reading filenames from '
                'disk will not work properly for filenames containing non-ascii characters. Make sure your '
                'locale env variables are set up correctly for the environment which is launching FlexGet.'
            )

    def _add_tray_icon_items(self, tray_icon: 'TrayIcon'):
        tray_icon.add_menu_item(text='Shutdown', action=self.shutdown, index=2)
        tray_icon.add_menu_item(text='Reload Config', action=self.load_config, index=3)
        tray_icon.add_menu_separator(index=4)

    @staticmethod
    def _init_options(args: List[str]) -> argparse.Namespace:
        """Initialize argument parsing"""
        try:
            options = CoreArgumentParser().parse_known_args(args, do_help=False)[0]
        except ParserError as exc:
            try:
                # If a non-built-in command was used, we need to parse with a parser that
                # doesn't define the subparsers
                options = manager_parser.parse_known_args(args, do_help=False)[0]
            except ParserError:
                manager_parser.print_help()
                print(f'\nError: {exc.message}')
                sys.exit(1)
        return options

    def _init_logging(self, to_file: bool = True) -> None:
        """Initialize logging facilities"""
        log_file = os.path.expanduser(self.options.logfile)
        # If an absolute path is not specified, use the config directory.
        if not os.path.isabs(log_file):
            log_file = os.path.join(self.config_base, log_file)
        self.log_filename = log_file
        flexget.log.start(
            log_file, self.options.loglevel, to_file=to_file, to_console=not self.options.cron
        )

    def initialize(self) -> None:
        """
        Load plugins, database, and config. Also initializes (but does not start) the task queue and ipc server.
        This should only be called after obtaining a lock.
        """
        if self.initialized:
            raise RuntimeError('Cannot call initialize on an already initialized manager.')

        plugin.load_plugins(
            extra_plugins=[os.path.join(self.config_base, 'plugins')],
            extra_components=[os.path.join(self.config_base, 'components')],
        )

        # Reparse CLI options now that plugins are loaded
        self.options = get_parser().parse_args(self.args)

        self.task_queue = TaskQueue()
        self.ipc_server = IPCServer(self, self.options.ipc_port)

        self.setup_yaml()
        self.init_sqlalchemy()
        fire_event('manager.initialize', self)
        try:
            self.load_config()
        except ValueError as e:
            logger.critical('Failed to load config file: {}', e.args[0])
            raise

        # cannot be imported at module level because of circular references
        from flexget.utils.simple_persistence import SimplePersistence

        self.persist = SimplePersistence('manager')

        if db_schema.upgrade_required():
            logger.info('Database upgrade is required. Attempting now.')
            fire_event('manager.upgrade', self)
            if manager.db_upgraded:
                fire_event('manager.db_upgraded', self)
        fire_event('manager.startup', self)
        self.initialized = True

    @property
    def tasks(self) -> List[str]:
        """A list of tasks in the config"""
        if not self.config:
            return []
        return list(self.config.get('tasks', {}).keys())

    @property
    def has_lock(self) -> bool:
        return self._has_lock

    def execute(
        self,
        options: Union[dict, argparse.Namespace] = None,
        priority: int = 1,
        suppress_warnings: Sequence[str] = None,
    ) -> List[Tuple[str, str, threading.Event]]:
        """
        Run all (can be limited with options) tasks from the config.

        :param options: Either an :class:`argparse.Namespace` instance, or a dict, containing options for execution
        :param priority: If there are other executions waiting to be run, they will be run in priority order,
            lowest first.
        :param suppress_warnings: Allows suppressing log warning about missing plugin in key phases
        :returns: a list of :class:`threading.Event` instances which will be
            set when each respective task has finished running
        """
        if options is None:
            options = copy.copy(self.options.execute)
        elif isinstance(options, dict):
            options_namespace = copy.copy(self.options.execute)
            options_namespace.__dict__.update(options)
            options = options_namespace
        task_names = self.tasks
        # Only reload config if daemon
        config_hash = self.hash_config()
        if self.is_daemon and self.autoreload_config and self.config_file_hash != config_hash:
            logger.info('Config change detected. Reloading.')
            try:
                self.load_config(output_to_console=False, config_file_hash=config_hash)
                logger.info('Config successfully reloaded!')
            except Exception as e:
                logger.error('Reloading config failed: {}', e)
        # Handle --tasks
        if options.tasks:
            # Consider * the same as not specifying tasks at all (makes sure manual plugin still works)
            if options.tasks == ['*']:
                options.tasks = None
            else:
                # Create list of tasks to run, preserving order
                task_names = []
                for arg in options.tasks:
                    matches = [
                        t for t in self.tasks if fnmatch.fnmatchcase(str(t).lower(), arg.lower())
                    ]
                    if not matches:
                        msg = f'`{arg}` does not match any tasks'
                        logger.error(msg)
                        continue
                    task_names.extend(m for m in matches if m not in task_names)
                # Set the option as a list of matching task names so plugins can use it easily
                options.tasks = task_names
        # TODO: 1.2 This is a hack to make task priorities work still, not sure if it's the best one
        task_names = sorted(
            task_names, key=lambda t: self.config['tasks'][t].get('priority', 65535)
        )

        finished_events = []
        for task_name in task_names:
            task = Task(
                self,
                task_name,
                options=options,
                output=get_console_output(),
                session_id=flexget.log.get_log_session_id(),
                priority=priority,
                suppress_warnings=suppress_warnings,
            )
            self.task_queue.put(task)
            finished_events.append((task.id, task.name, task.finished_event))
        return finished_events

    def start(self) -> None:
        """
        Starting point when executing from commandline, dispatch execution to correct destination.

        If there is a FlexGet process with an ipc server already running, the command will be sent there for execution
        and results will be streamed back.
        If not, this will attempt to obtain a lock, initialize the manager, and run the command here.
        """
        # When we are in test mode, we use a different lock file and db
        if self.options.test:
            self.lockfile = os.path.join(self.config_base, f'.test-{self.config_name}-lock')
        # If another process is started, send the execution to the running process
        ipc_info = self.check_ipc_info()
        # If we are connecting to a running daemon, we don't want to log to the log file,
        # the daemon is already handling that.
        self._init_logging(to_file=not ipc_info)
        if ipc_info:
            console(
                'There is a FlexGet process already running for this config, sending execution there.'
            )
            logger.debug('Sending command to running FlexGet process: {}', self.args)
            try:
                client = IPCClient(ipc_info['port'], ipc_info['password'])
            except ValueError as e:
                logger.error(e)
            else:
                try:
                    client.handle_cli(self.args)
                except KeyboardInterrupt:
                    logger.error(
                        'Disconnecting from daemon due to ctrl-c. Executions will still continue in the '
                        'background.'
                    )
                except EOFError:
                    logger.error('Connection from daemon was severed.')
            return
        if self.options.test:
            logger.info('Test mode, creating a copy from database ...')
            db_test_filename = os.path.join(self.config_base, f'test-{self.config_name}.sqlite')
            if os.path.exists(self.db_filename):
                shutil.copy(self.db_filename, db_test_filename)
                logger.info('Test database created')
            self.db_filename = db_test_filename
        # No running process, we start our own to handle command
        with self.acquire_lock():
            self.initialize()
            self.handle_cli()
            self._shutdown()

    def handle_cli(self, options: argparse.Namespace = None) -> None:
        """
        Dispatch a cli command to the appropriate function.

        * :meth:`.execute_command`
        * :meth:`.daemon_command`
        * CLI plugin callback function

        The manager should have a lock and be initialized before calling this method.

        :param options: argparse options for command. Defaults to options that manager was instantiated with.
        """
        if not options:
            options = self.options
        command = options.cli_command
        if command is None:
            raise Exception('Command missing')
        command_options = getattr(options, command)
        # First check for built-in commands
        if command in ['execute', 'daemon']:
            if command == 'execute':
                self.execute_command(command_options)
            elif command == 'daemon':
                self.daemon_command(command_options)
        else:
            # Otherwise dispatch the command to the callback function
            options.cli_command_callback(self, command_options)

    def execute_command(self, options: argparse.Namespace) -> None:
        """
        Handles the 'execute' CLI command.

        If there is already a task queue running in this process, adds the execution to the queue.
        If FlexGet is being invoked with this command, starts up a task queue and runs the execution.

        Fires events:

        * manager.execute.started
        * manager.execute.completed

        :param options: argparse options
        """
        fire_event('manager.execute.started', self, options)
        if self.task_queue.is_alive() or self.is_daemon:
            if not self.task_queue.is_alive():
                logger.error(
                    'Task queue has died unexpectedly. Restarting it. Please open an issue on Github and include'
                    ' any previous error logs.'
                )
                self.task_queue = TaskQueue()
                self.task_queue.start()
            if len(self.task_queue):
                logger.verbose('There is a task already running, execution queued.')
            finished_events = self.execute(options)
            if not options.cron:
                # Wait until execution of all tasks has finished
                for _, _, event in finished_events:
                    event.wait()
        else:
            self.task_queue.start()
            self.ipc_server.start()
            self.execute(options)
            self.shutdown(finish_queue=True)
            self.task_queue.wait()
        fire_event('manager.execute.completed', self, options)

    def daemon_command(self, options: argparse.Namespace) -> None:
        """
        Handles the 'daemon' CLI command.

        Fires events:

        * manager.daemon.started
        * manager.daemon.completed

        :param options: argparse options
        """

        # Import API so it can register to daemon.started event
        if options.action == 'start':
            if self.is_daemon:
                logger.error('Daemon already running for this config.')
                return
            elif self.task_queue.is_alive():
                logger.error(
                    'Non-daemon execution of FlexGet is running. Cannot start daemon until it is finished.'
                )
                return
            if options.daemonize:
                self.daemonize()
            if options.autoreload_config:
                self.autoreload_config = True
            try:
                signal.signal(signal.SIGTERM, self._handle_sigterm)
            except ValueError as e:
                # If flexget is being called from another script, e.g. windows service helper, and we are not the
                # main thread, this error will occur.
                logger.debug('Error registering sigterm handler: {}', e)
            self.is_daemon = True

            def run_daemon(tray_icon: 'TrayIcon' = None):
                fire_event('manager.daemon.started', self)
                self.task_queue.start()
                self.ipc_server.start()
                self.task_queue.wait()
                fire_event('manager.daemon.completed', self)
                if tray_icon:
                    tray_icon.stop()

            if options.tray_icon:
                from flexget.tray_icon import tray_icon  # noqa

                self._add_tray_icon_items(tray_icon)

                # Tray icon must be run in the main thread.
                m = threading.Thread(target=run_daemon, args=(tray_icon,))
                m.start()
                tray_icon.run()
                m.join()
            else:
                run_daemon()

        elif options.action in ['stop', 'reload-config', 'status']:
            if not self.is_daemon:
                console('There does not appear to be a daemon running.')
                return
            if options.action == 'status':
                logger.debug('`daemon status` called. Daemon running. (PID: {})', os.getpid())
                console(f'Daemon running. (PID: {os.getpid()})')
            elif options.action == 'stop':
                tasks = (
                    'all queued tasks (if any) have'
                    if options.wait
                    else 'currently running task (if any) has'
                )
                logger.info(
                    'Daemon shutdown requested. Shutdown will commence when {} finished executing.',
                    tasks,
                )
                self.shutdown(options.wait)
            elif options.action == 'reload-config':
                logger.info('Reloading config from disk.')
                try:
                    self.load_config()
                except ValueError as e:
                    logger.error('Error loading config: {}', e.args[0])
                else:
                    logger.info('Config successfully reloaded from disk.')

    def _handle_sigterm(self, signum, frame) -> None:
        logger.info('Got SIGTERM. Shutting down.')
        self.shutdown(finish_queue=False)

    def setup_yaml(self) -> None:
        """Sets up the yaml loader to return unicode objects for strings by default"""

        def construct_yaml_str(self, node):
            # Override the default string handling function
            # to always return unicode objects
            return self.construct_scalar(node)

        yaml.Loader.add_constructor('tag:yaml.org,2002:str', construct_yaml_str)
        yaml.SafeLoader.add_constructor('tag:yaml.org,2002:str', construct_yaml_str)

        # Set up the dumper to not tag every string with !!python/unicode
        def unicode_representer(dumper, uni):
            node = yaml.ScalarNode(tag='tag:yaml.org,2002:str', value=uni)
            return node

        yaml.add_representer(str, unicode_representer)

        # Set up the dumper to increase the indent for lists
        def increase_indent_wrapper(func):
            def increase_indent(self, flow=False, indentless=False):
                func(self, flow, False)

            return increase_indent

        yaml.Dumper.increase_indent = increase_indent_wrapper(yaml.Dumper.increase_indent)
        yaml.SafeDumper.increase_indent = increase_indent_wrapper(yaml.SafeDumper.increase_indent)

    def _init_config(self, create: bool = False) -> None:
        """
        Find and load the configuration file.

        :param bool create: If a config file is not found, and create is True, one will be created in the home folder
        :raises: `OSError` when no config file could be found, and `create` is False.
        """
        home_path = os.path.join(os.path.expanduser('~'), '.flexget')
        options_config = os.path.expanduser(self.options.config)

        possible = []
        if os.path.isabs(options_config):
            # explicit path given, don't try anything
            config = options_config
            possible = [config]
        else:
            logger.debug('Figuring out config load paths')
            try:
                possible.append(os.getcwd())
            except OSError:
                logger.debug('current directory invalid, not searching for config there')
            # for virtualenv / dev sandbox
            if hasattr(sys, 'real_prefix'):
                logger.debug('Adding virtualenv path')
                possible.append(sys.prefix)
            # normal lookup locations
            possible.append(home_path)
            if sys.platform.startswith('win'):
                # On windows look in ~/flexget as well, as explorer does not let you create a folder starting with a dot
                home_path = os.path.join(os.path.expanduser('~'), 'flexget')
                possible.append(home_path)
            else:
                # The freedesktop.org standard config location
                xdg_config = os.environ.get(
                    'XDG_CONFIG_HOME', os.path.join(os.path.expanduser('~'), '.config')
                )
                possible.append(os.path.join(xdg_config, 'flexget'))

            for path in possible:
                config = os.path.join(path, options_config)
                if os.path.exists(config):
                    logger.debug('Found config: {}', config)
                    break
            else:
                config = None

        if create and not (config and os.path.exists(config)):
            config = os.path.join(home_path, options_config)
            logger.info('Config file {} not found. Creating new config {}', options_config, config)
            with open(config, 'w') as newconfig:
                # Write empty tasks to the config
                newconfig.write(yaml.dump({'tasks': {}}))
        elif not config:
            logger.critical('Failed to find configuration file {}', options_config)
            logger.info('Tried to read from: {}', ', '.join(possible))
            raise OSError('No configuration file found.')
        if not os.path.isfile(config):
            raise OSError(f'Config `{config}` does not appear to be a file.')

        logger.debug('Config file {} selected', config)
        self.config_path = config
        self.config_name = os.path.splitext(os.path.basename(config))[0]
        self.config_base = os.path.normpath(os.path.dirname(config))
        self.lockfile = os.path.join(self.config_base, f'.{self.config_name}-lock')
        self.db_filename = os.path.join(self.config_base, f'db-{self.config_name}.sqlite')

    def hash_config(self) -> Optional[str]:
        if not self.config_path:
            return None
        sha1_hash = hashlib.sha1()
        with open(self.config_path, 'rb') as f:
            while True:
                data = f.read(65536)
                if not data:
                    break
                sha1_hash.update(data)
        return sha1_hash.hexdigest()

    def load_config(self, output_to_console: bool = True, config_file_hash: str = None) -> None:
        """
        Loads the config file from disk, validates and activates it.

        :raises: `ValueError` if there is a problem loading the config file
        """
        fire_event('manager.before_config_load', self)
        with open(self.config_path, 'r', encoding='utf-8') as f:
            try:
                raw_config = f.read()
            except UnicodeDecodeError:
                logger.critical('Config file must be UTF-8 encoded.')
                raise ValueError('Config file is not UTF-8 encoded')
        try:
            self.config_file_hash = config_file_hash or self.hash_config()
            config = yaml.safe_load(raw_config) or {}
        except yaml.YAMLError as e:
            msg = str(e).replace('\n', ' ')
            msg = ' '.join(msg.split())
            logger.critical(msg)
            if output_to_console:
                print('')
                print('-' * 79)
                print(' Malformed configuration file (check messages above). Common reasons:')
                print('-' * 79)
                print('')
                print(' o Indentation error')
                print(' o Missing : from end of the line')
                print(' o Non ASCII characters (use UTF8)')
                print(
                    ' o If text contains any of :[]{}% characters it must be single-quoted '
                    '(eg. value{1} should be \'value{1}\')\n'
                )

                # Not very good practice but we get several kind of exceptions here, I'm not even sure all of them
                # At least: ReaderError, YmlScannerError (or something like that)
                if isinstance(e, yaml.MarkedYAMLError):
                    lines = 0
                    if e.problem is not None:
                        print(f' Reason: {e.problem}\n')
                        if e.problem == 'mapping values are not allowed here':
                            print(' ----> MOST LIKELY REASON: Missing : from end of the line!')
                            print('')
                    if e.context_mark is not None:
                        print(
                            ' Check configuration near line %s, column %s'
                            % (e.context_mark.line, e.context_mark.column)
                        )
                        lines += 1
                    if e.problem_mark is not None:
                        print(
                            ' Check configuration near line %s, column %s'
                            % (e.problem_mark.line, e.problem_mark.column)
                        )
                        lines += 1
                    if lines:
                        print('')
                    if lines == 1:
                        print(' Fault is almost always in this or previous line\n')
                    if lines == 2:
                        print(' Fault is almost always in one of these lines or previous ones\n')

            # When --debug escalate to full stacktrace
            if self.options.debug or not output_to_console:
                raise
            raise ValueError('Config file is not valid YAML')

        # config loaded successfully
        logger.debug('config_name: {}', self.config_name)
        logger.debug('config_base: {}', self.config_base)
        # Install the newly loaded config
        self.update_config(config)

    def update_config(self, config: dict) -> None:
        """
        Provide a new config for the manager to use.

        :raises: `ValueError` and rolls back to previous config if the provided config is not valid.
        """
        new_user_config = config
        old_config = self.config
        try:
            self.config = self.validate_config(config)
        except ConfigError as e:
            for error in getattr(e, 'errors', []):
                logger.critical('[{}] {}', error.json_pointer, error.message)
            logger.debug('invalid config, rolling back')
            self.config = old_config
            raise
        logger.debug('New config data loaded.')
        self.user_config = copy.deepcopy(new_user_config)
        fire_event('manager.config_updated', self)

    def backup_config(self) -> str:
        backup_path = os.path.join(
            self.config_base,
            f'{self.config_name}-{datetime.now().strftime("%y%m%d%H%M%S")}.bak',
        )

        logger.debug(f'backing up old config to {backup_path} before new save')
        try:
            shutil.copy(self.config_path, backup_path)
        except OSError as e:
            logger.warning(f'Config backup creation failed: {e}')
            raise
        return backup_path

    def save_config(self) -> None:
        """Dumps current config to yaml config file"""
        # TODO: Only keep x number of backups..

        # Back up the user's current config before overwriting
        try:
            self.backup_config()
        except OSError:
            return
        with open(self.config_path, 'w') as config_file:
            config_file.write(yaml.dump(self.user_config, default_flow_style=False))

    def config_changed(self) -> None:
        """Makes sure that all tasks will have the config_modified flag come out true on the next run.
        Useful when changing the db and all tasks need to be completely reprocessed."""
        from flexget.task import config_changed

        config_changed()
        fire_event('manager.config_updated', self)

    def validate_config(self, config: dict = None) -> dict:
        """
        Check all root level keywords are valid. Config may be modified by before_config_validate hooks. Modified
        config will be returned.

        :param config: Config to check. If not provided, current manager config will be checked.
        :raises: `ValueError` when config fails validation. There will be an `errors` attribute with the schema errors.
        :returns: Final validated config.
        """
        conf = config if config else self.config
        conf = fire_event('manager.before_config_validate', conf, self)
        errors = config_schema.process_config(conf)
        if errors:
            err = ConfigError('Did not pass schema validation.')
            err.errors = errors
            raise err
        else:
            return conf

    def init_sqlalchemy(self) -> None:
        """Initialize SQLAlchemy"""
        try:
            if [int(part) for part in sqlalchemy.__version__.split('.')] < [0, 7, 0]:
                print(
                    'FATAL: SQLAlchemy 0.7.0 or newer required. Please upgrade your SQLAlchemy.',
                    file=sys.stderr,
                )
                sys.exit(1)
        except ValueError:
            logger.critical('Failed to check SQLAlchemy version, you may need to upgrade it')

        # SQLAlchemy
        if not self.database_uri:
            # in case running on windows, needs double \\
            filename = self.db_filename.replace('\\', '\\\\')
            self.database_uri = f'sqlite:///{filename}'

        if self.db_filename and not os.path.exists(self.db_filename):
            logger.verbose('Creating new database {} - DO NOT INTERRUPT ...', self.db_filename)

        # fire up the engine
        logger.debug('Connecting to: {}', self.database_uri)
        try:
            self.engine = sqlalchemy.create_engine(
                self.database_uri,
                echo=self.options.debug_sql,
                connect_args={'check_same_thread': False, 'timeout': 10},
            )
        except ImportError as e:
            print(
                'FATAL: Unable to use SQLite. Are you running Python 2.7, 3.3 or newer ?\n'
                'Python should normally have SQLite support built in.\n'
                'If you\'re running correct version of Python then it is not equipped with SQLite.\n'
                'You can try installing `pysqlite`. If you have compiled python yourself, '
                'recompile it with SQLite support.\n'
                f'Error: {e}',
                file=sys.stderr,
            )
            sys.exit(1)
        Session.configure(bind=self.engine)
        # create all tables, doesn't do anything to existing tables
        try:
            Base.metadata.create_all(bind=self.engine)
        except OperationalError as e:
            if os.path.exists(self.db_filename):
                print(
                    '%s - make sure you have write permissions to file %s'
                    % (e.message, self.db_filename),
                    file=sys.stderr,
                )
            else:
                print(
                    '%s - make sure you have write permissions to directory %s'
                    % (e.message, self.config_base),
                    file=sys.stderr,
                )
            raise

    def _read_lock(self) -> Optional[dict]:
        """
        Read the values from the lock file. Returns None if there is no current lock file.
        """
        if self.lockfile and os.path.exists(self.lockfile):
            result: Dict[str, Union[str, int]] = {}
            with open(self.lockfile, encoding='utf-8') as f:
                lines = [line for line in f.readlines() if line]
            for line in lines:
                try:
                    key, value = line.split(':', 1)
                except ValueError:
                    logger.debug('Invalid line in lock file: {}', line)
                    continue
                result[key.strip().lower()] = value.strip()
            for key in result:
                if result[key].isdigit():
                    result[key] = int(result[key])
            result.setdefault('pid', None)
            if not result['pid']:
                logger.error(
                    'Invalid lock file. Make sure FlexGet is not running, then delete it.'
                )
            elif not pid_exists(result['pid']):
                return None
            return result
        return None

    def check_lock(self) -> bool:
        """Returns True if there is a lock on the database."""
        lock_info = self._read_lock()
        if not lock_info:
            return False
        # Don't count it if we hold the lock
        if os.getpid() == lock_info['pid']:
            return False
        return True

    def check_ipc_info(self) -> Optional[dict]:
        """If a daemon has a lock on the database, return info to connect to IPC."""
        lock_info = self._read_lock()
        if lock_info and 'port' in lock_info:
            return lock_info
        return None

    @contextmanager
    def acquire_lock(self, event: bool = True) -> Iterator:
        """
        :param bool event: If True, the 'manager.lock_acquired' event will be fired after a lock is obtained
        """
        acquired = False
        try:
            # Don't do anything if we already have a lock. This means only the outermost call will release the lock file
            if not self._has_lock:
                # Exit if there is an existing lock.
                if self.check_lock():
                    with open(self.lockfile, encoding='utf-8') as f:
                        pid = f.read()
                    print(
                        'Another process (%s) is running, will exit.' % pid.split('\n')[0],
                        file=sys.stderr,
                    )
                    print(
                        "If you're sure there is no other instance running, delete %s"
                        % self.lockfile,
                        file=sys.stderr,
                    )
                    sys.exit(1)

                self._has_lock = True
                self.write_lock()
                acquired = True
                if event:
                    fire_event('manager.lock_acquired', self)
            yield
        finally:
            if acquired:
                self.release_lock()
                self._has_lock = False

    def write_lock(self, ipc_info: dict = None) -> None:
        assert self._has_lock
        with open(self.lockfile, 'w', encoding='utf-8') as f:
            f.write(f'PID: {os.getpid()}\n')
            if ipc_info:
                for key in sorted(ipc_info):
                    f.write(f'{key}: {ipc_info[key]}\n')

    def release_lock(self) -> None:
        try:
            os.remove(self.lockfile)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
            logger.debug(f'Lockfile {self.lockfile} not found')
        else:
            logger.debug(f'Removed {self.lockfile}')

    def daemonize(self) -> None:
        """Daemonizes the current process. Returns the new pid"""
        if sys.platform.startswith('win'):
            logger.error('Cannot daemonize on windows')
            return
        if threading.active_count() != 1:
            logger.critical(
                'There are {!r} active threads. Daemonizing now may cause strange failures.',
                threading.enumerate(),
            )

        logger.info('Daemonizing...')

        try:
            pid = os.fork()
            if pid > 0:
                # Don't run the exit handlers on the parent
                atexit._exithandlers = []
                # exit first parent
                sys.exit(0)
        except OSError as e:
            sys.stderr.write(f'fork #1 failed: {e.errno} ({e.strerror})\n')
            sys.exit(1)

        # decouple from parent environment
        os.chdir('/')
        os.setsid()
        os.umask(0)

        # do second fork
        try:
            pid = os.fork()
            if pid > 0:
                # Don't run the exit handlers on the parent
                atexit._exithandlers = []
                # exit from second parent
                sys.exit(0)
        except OSError as e:
            sys.stderr.write(f'fork #2 failed: {e.errno} ({e.strerror})\n')
            sys.exit(1)

        logger.info(f'Daemonize complete. New PID: {os.getpid()}')
        # redirect standard file descriptors
        sys.stdout.flush()
        sys.stderr.flush()
        si = open(os.devnull, 'r')
        so = open(os.devnull, 'ab+')
        se = open(os.devnull, 'ab+', 0)
        os.dup2(si.fileno(), sys.stdin.fileno())
        os.dup2(so.fileno(), sys.stdout.fileno())
        os.dup2(se.fileno(), sys.stderr.fileno())

        # If we have a lock, update the lock file with our new pid
        if self._has_lock:
            self.write_lock()

    def db_cleanup(self, force: bool = False) -> None:
        """
        Perform database cleanup if cleanup interval has been met.

        Fires events:

        * manager.db_cleanup

          If interval was met. Gives session to do the cleanup as a parameter.

        :param bool force: Run the cleanup no matter whether the interval has been met.
        """
        expired = (
            self.persist.get('last_cleanup', datetime(1900, 1, 1))
            < datetime.now() - DB_CLEANUP_INTERVAL
        )
        if force or expired:
            logger.info('Running database cleanup.')
            with Session() as session:
                fire_event('manager.db_cleanup', self, session)
            # Try to VACUUM after cleanup
            fire_event('manager.db_vacuum', self)
            # Just in case some plugin was overzealous in its cleaning, mark the config changed
            self.config_changed()
            self.persist['last_cleanup'] = datetime.now()
        else:
            logger.debug('Not running db cleanup, last run {}', self.persist.get('last_cleanup'))

    def shutdown(self, finish_queue: bool = True) -> None:
        """
        Request manager shutdown.

        :param bool finish_queue: Should scheduler finish the task queue
        """
        if not self.initialized:
            raise RuntimeError('Cannot shutdown manager that was never initialized.')
        fire_event('manager.shutdown_requested', self)
        self.task_queue.shutdown(finish_queue)

    def _shutdown(self) -> None:
        """Runs when the manager is done processing everything."""
        if self.ipc_server:
            self.ipc_server.shutdown()
        fire_event('manager.shutdown', self)
        if not self.unit_test:  # don't scroll "nosetests" summary results when logging is enabled
            logger.debug('Shutting down')
        self.engine.dispose()
        # remove temporary database used in test mode
        if self.options.test:
            if 'test' not in self.db_filename:
                raise Exception('trying to delete non test database?')
            if self._has_lock:
                os.remove(self.db_filename)
                logger.info('Removed test database')
        global manager
        manager = None

    def crash_report(self) -> str:
        """
        This should be called when handling an unexpected exception. Will create a new log file containing the last 50
        debug messages as well as the crash traceback.
        """
        if not self.unit_test:
            log_dir = os.path.dirname(self.log_filename)
            filename = os.path.join(
                log_dir, datetime.now().strftime('crash_report.%Y.%m.%d.%H%M%S%f.log')
            )
            with codecs.open(filename, 'w', encoding='utf-8') as outfile:
                outfile.writelines(flexget.log.debug_buffer)
                traceback.print_exc(file=outfile)
            logger.critical(
                'An unexpected crash has occurred. Writing crash report to {}. '
                'Please verify you are running the latest version of flexget by using "flexget -V" '
                'from CLI or by using version_checker plugin'
                ' at http://flexget.com/wiki/Plugins/version_checker. '
                'You are currently using version {}',
                filename,
                get_current_flexget_version(),
            )
        logger.opt(exception=True).debug('Traceback:')
        return traceback.format_exc()



from loguru import logger

from flexget import options, plugin
from flexget.event import event
from flexget.terminal import console

logger = logger.bind(name='dump')


def dump(entries, debug=False, eval_lazy=False, trace=False, title_only=False):
    """
    Dump *entries* to stdout

    :param list entries: Entries to be dumped.
    :param bool debug: Print non printable fields as well.
    :param bool eval_lazy: Evaluate lazy fields.
    :param bool trace: Display trace information.
    :param bool title_only: Display only title field
    """

    def sort_key(field):
        # Sort certain fields above the rest
        if field == 'title':
            return (0,)
        if field == 'url':
            return (1,)
        if field == 'original_url':
            return (2,)
        return 3, field

    for entry in entries:
        for field in sorted(entry, key=sort_key):
            if field.startswith('_') and not debug:
                continue
            if title_only and field != 'title':
                continue
            if entry.is_lazy(field) and not eval_lazy:
                value = '<LazyField - value will be determined when it is accessed>'
            else:
                try:
                    value = entry[field]
                except KeyError:
                    value = '<LazyField - lazy lookup failed>'
            if isinstance(value, str):
                try:
                    console('%-17s: %s' % (field, value.replace('\r', '').replace('\n', '')))
                except Exception:
                    console('%-17s: %r (warning: unable to print)' % (field, value))
            elif isinstance(value, list):
                console('%-17s: %s' % (field, '[%s]' % ', '.join(str(v) for v in value)))
            elif isinstance(value, (int, float, dict)):
                console('%-17s: %s' % (field, value))
            elif value is None:
                console('%-17s: %s' % (field, value))
            else:
                try:
                    value = str(entry[field])
                    console('%-17s: %s' % (field, value.replace('\r', '').replace('\n', '')))
                except Exception:
                    if debug:
                        console('%-17s: [not printable] (%r)' % (field, value))
        if trace:
            console('-- Processing trace:')
            for item in entry.traces:
                console('%-10s %-7s %s' % (item[0], '' if item[1] is None else item[1], item[2]))
        if not title_only:
            console('')


class OutputDump:
    """
    Outputs all entries to console
    """

    schema = {'type': 'boolean'}

    @plugin.priority(0)
    def on_task_output(self, task, config):
        if not config and task.options.dump_entries is None:
            return

        eval_lazy = 'eval' in task.options.dump_entries
        trace = 'trace' in task.options.dump_entries
        title = 'title' in task.options.dump_entries
        states = ['accepted', 'rejected', 'failed', 'undecided']
        dumpstates = [s for s in states if s in task.options.dump_entries]
        specificstates = dumpstates
        if not dumpstates:
            dumpstates = states
        undecided = [entry for entry in task.all_entries if entry.undecided]
        if 'undecided' in dumpstates:
            if undecided:
                console('-- Undecided: --------------------------')
                dump(undecided, task.options.debug, eval_lazy, trace, title)
            elif specificstates:
                console('No undecided entries')
        if 'accepted' in dumpstates:
            if task.accepted:
                console('-- Accepted: ---------------------------')
                dump(task.accepted, task.options.debug, eval_lazy, trace, title)
            elif specificstates:
                console('No accepted entries')
        if 'rejected' in dumpstates:
            if task.rejected:
                console('-- Rejected: ---------------------------')
                dump(task.rejected, task.options.debug, eval_lazy, trace, title)
            elif specificstates:
                console('No rejected entries')
        if 'failed' in dumpstates:
            if task.failed:
                console('-- Failed: -----------------------------')
                dump(task.failed, task.options.debug, eval_lazy, trace, title)
            elif specificstates:
                console('No failed entries')


@event('plugin.register')
def register_plugin():
    plugin.register(OutputDump, 'dump', builtin=True, api_ver=2)


@event('options.register')
def register_parser_arguments():
    options.get_parser('execute').add_argument(
        '--dump',
        nargs='*',
        choices=['eval', 'trace', 'accepted', 'rejected', 'undecided', 'title'],
        dest='dump_entries',
        help=(
            'display all entries in task with fields they contain, '
            'use `--dump eval` to evaluate all lazy fields. Specify an entry '
            'state/states to only dump matching entries.'
        ),
    )



import difflib
import time
from datetime import datetime, timedelta
from typing import Any, Dict, Optional, Type
from urllib.error import URLError
from urllib.parse import quote_plus

from loguru import logger
from sqlalchemy import Column, DateTime, Integer, String, Table, func, sql
from sqlalchemy.orm import Session, relation
from sqlalchemy.schema import ForeignKey, Index

from flexget import db_schema, plugin
from flexget.plugin import PluginError, internet
from flexget.utils import requests
from flexget.utils.database import text_date_synonym, with_session
from flexget.utils.sqlalchemy_utils import table_add_column, table_schema

logger = logger.bind(name='api_rottentomatoes')
Base: Type[db_schema.VersionedBaseMeta] = db_schema.versioned_base('api_rottentomatoes', 2)
session = requests.Session()
# There is a 5 call per second rate limit per api key with multiple users on the same api key, this can be problematic
session.add_domain_limiter(requests.TimedLimiter('api.rottentomatoes.com', '0.4 seconds'))

# This is developer Atlanta800's API key
API_KEY = 'rh8chjzp8vu6gnpwj88736uv'
API_VER = 'v1.0'
SERVER = 'http://api.rottentomatoes.com/api/public'

MIN_MATCH = 0.5
MIN_DIFF = 0.01


@db_schema.upgrade('api_rottentomatoes')
def upgrade(ver: int, session: Session) -> int:
    if ver == 0:
        table_names = [
            'rottentomatoes_actors',
            'rottentomatoes_alternate_ids',
            'rottentomatoes_directors',
            'rottentomatoes_genres',
            'rottentomatoes_links',
            'rottentomatoes_movie_actors',
            'rottentomatoes_movie_directors',
            'rottentomatoes_movie_genres',
            'rottentomatoes_movies',
            'rottentomatoes_posters',
            'rottentomatoes_releasedates',
            'rottentomatoes_search_results',
        ]
        tables = [table_schema(name, session) for name in table_names]
        for table in tables:
            session.execute(table.delete())
        table_add_column('rottentomatoes_actors', 'rt_id', String, session)
        ver = 1
    if ver == 1:
        table = table_schema('rottentomatoes_search_results', session)
        session.execute(sql.delete(table, table.c.movie_id == None))
        ver = 2
    return ver


# association tables
genres_table = Table(
    'rottentomatoes_movie_genres',
    Base.metadata,
    Column('movie_id', Integer, ForeignKey('rottentomatoes_movies.id')),
    Column('genre_id', Integer, ForeignKey('rottentomatoes_genres.id')),
    Index('ix_rottentomatoes_movie_genres', 'movie_id', 'genre_id'),
)
Base.register_table(genres_table)

actors_table = Table(
    'rottentomatoes_movie_actors',
    Base.metadata,
    Column('movie_id', Integer, ForeignKey('rottentomatoes_movies.id')),
    Column('actor_id', Integer, ForeignKey('rottentomatoes_actors.id')),
    Index('ix_rottentomatoes_movie_actors', 'movie_id', 'actor_id'),
)
Base.register_table(actors_table)

directors_table = Table(
    'rottentomatoes_movie_directors',
    Base.metadata,
    Column('movie_id', Integer, ForeignKey('rottentomatoes_movies.id')),
    Column('director_id', Integer, ForeignKey('rottentomatoes_directors.id')),
    Index('ix_rottentomatoes_movie_directors', 'movie_id', 'director_id'),
)
Base.register_table(directors_table)


# TODO: get rid of
class RottenTomatoesContainer:
    """Base class for RottenTomatoes objects"""

    def __init__(self, init_dict: Optional[Dict[str, Any]] = None) -> None:
        if isinstance(init_dict, dict):
            self.update_from_dict(init_dict)

    def update_from_dict(self, update_dict: Dict[str, Any]) -> None:
        """Populates any simple (string or number) attributes from a dict"""
        for col in self.__table__.columns:
            if isinstance(update_dict.get(col.name), (str, int, float)):
                setattr(self, col.name, update_dict[col.name])


class RottenTomatoesMovie(RottenTomatoesContainer, Base):
    __tablename__ = 'rottentomatoes_movies'

    id = Column(Integer, primary_key=True, autoincrement=False, nullable=False)
    title = Column(String)
    year = Column(Integer)
    genres = relation('RottenTomatoesGenre', secondary=genres_table, backref='movies')
    mpaa_rating = Column(String)
    runtime = Column(Integer)
    critics_consensus = Column(String)
    release_dates = relation('ReleaseDate', backref='movie', cascade='all, delete, delete-orphan')
    critics_rating = Column(String)
    critics_score = Column(Integer)
    audience_rating = Column(String)
    audience_score = Column(Integer)
    synopsis = Column(String)
    posters = relation(
        'RottenTomatoesPoster', backref='movie', cascade='all, delete, delete-orphan'
    )
    cast = relation('RottenTomatoesActor', secondary=actors_table, backref='movies')
    directors = relation('RottenTomatoesDirector', secondary=directors_table, backref='movies')
    studio = Column(String)
    # NOTE: alternate_ids is not anymore used, it used to store imdb_id
    alternate_ids = relation(
        'RottenTomatoesAlternateId', backref='movie', cascade='all, delete, delete-orphan'
    )
    links = relation('RottenTomatoesLink', backref='movie', cascade='all, delete, delete-orphan')

    # updated time, so we can grab new rating counts after 48 hours
    # set a default, so existing data gets updated with a rating
    updated = Column(DateTime)

    @property
    def expired(self) -> bool:
        """
        :return: True if movie details are considered to be expired, ie. need of update
        """
        if self.updated is None:
            logger.debug('updated is None: {}', self)
            return True
        refresh_interval = 2
        if self.year:
            age = datetime.now().year - self.year
            refresh_interval += age * 5
            logger.debug('movie `{}` age {} expires in {} days', self.title, age, refresh_interval)
        return self.updated < datetime.now() - timedelta(days=refresh_interval)

    def __repr__(self) -> str:
        return '<RottenTomatoesMovie(title=%s,id=%s,year=%s)>' % (self.title, self.id, self.year)


class RottenTomatoesGenre(Base):
    __tablename__ = 'rottentomatoes_genres'

    id = Column(Integer, primary_key=True)
    name = Column(String)

    def __init__(self, name: str) -> None:
        self.name = name


class ReleaseDate(Base):
    __tablename__ = 'rottentomatoes_releasedates'

    db_id = Column(Integer, primary_key=True)
    movie_id = Column(Integer, ForeignKey('rottentomatoes_movies.id'))
    name = Column(String)
    date = text_date_synonym('_date')
    _date = Column('date', DateTime)

    def __init__(self, name: str, date: datetime) -> None:
        self.name = name
        self.date = date


class RottenTomatoesPoster(Base):
    __tablename__ = 'rottentomatoes_posters'

    db_id = Column(Integer, primary_key=True)
    movie_id = Column(Integer, ForeignKey('rottentomatoes_movies.id'))
    name = Column(String)
    url = Column(String)

    def __init__(self, name: str, url: str) -> None:
        self.name = name
        self.url = url


class RottenTomatoesActor(Base):
    __tablename__ = 'rottentomatoes_actors'

    id = Column(Integer, primary_key=True)
    rt_id = Column(String)
    name = Column(String)

    def __init__(self, name: str, rt_id: str) -> None:
        self.name = name
        self.rt_id = rt_id


class RottenTomatoesDirector(Base):
    __tablename__ = 'rottentomatoes_directors'

    id = Column(Integer, primary_key=True)
    name = Column(String)

    def __init__(self, name: str) -> None:
        self.name = name


class RottenTomatoesAlternateId(Base):
    __tablename__ = 'rottentomatoes_alternate_ids'

    db_id = Column(Integer, primary_key=True)
    movie_id = Column(Integer, ForeignKey('rottentomatoes_movies.id'))
    name = Column(String)
    id = Column(String)

    def __init__(self, name: str, id: str) -> None:
        self.name = name
        self.id = id


class RottenTomatoesLink(Base):
    __tablename__ = 'rottentomatoes_links'

    db_id = Column(Integer, primary_key=True)
    movie_id = Column(Integer, ForeignKey('rottentomatoes_movies.id'))
    name = Column(String)
    url = Column(String)

    def __init__(self, name: str, url: str) -> None:
        self.name = name
        self.url = url


class RottenTomatoesSearchResult(Base):
    __tablename__ = 'rottentomatoes_search_results'

    id = Column(Integer, primary_key=True)
    search = Column(String, nullable=False)
    movie_id = Column(Integer, ForeignKey('rottentomatoes_movies.id'), nullable=False)
    movie = relation(RottenTomatoesMovie, backref='search_strings')

    def __repr__(self) -> str:
        return '<RottenTomatoesSearchResult(search=%s,movie_id=%s,movie=%s)>' % (
            self.search,
            self.movie_id,
            self.movie,
        )


@internet(logger)
@with_session
def lookup_movie(
    title: Optional[str] = None,
    year: Optional[int] = None,
    rottentomatoes_id: Optional[int] = None,
    smart_match: Optional[bool] = None,
    only_cached: bool = False,
    session: Optional[Session] = None,
    api_key: Optional[str] = None,
) -> RottenTomatoesMovie:
    """
    Do a lookup from Rotten Tomatoes for the movie matching the passed arguments.
    Any combination of criteria can be passed, the most specific criteria specified will be used.

    :param rottentomatoes_id: rottentomatoes_id of desired movie
    :param string title: title of desired movie
    :param year: release year of desired movie
    :param smart_match: attempt to clean and parse title and year from a string
    :param only_cached: if this is specified, an online lookup will not occur if the movie is not in the cache
    :param session: optionally specify a session to use, if specified, returned Movie will be live in that session
    :param api_key: optionaly specify an API key to use
    :returns: The Movie object populated with data from Rotten Tomatoes
    :raises: PluginError if a match cannot be found or there are other problems with the lookup
    """
    if smart_match:
        # If smart_match was specified, and we don't have more specific criteria, parse it into a title and year
        title_parser = plugin.get('parsing', 'api_rottentomatoes').parse_movie(smart_match)
        title = title_parser.name
        year = title_parser.year
        if title == '' and not (rottentomatoes_id or title):
            raise PluginError('Failed to parse name from %s' % smart_match)

    search_string = ""
    if title:
        search_string = title.lower()
        if year:
            search_string = '%s %s' % (search_string, year)
    elif not rottentomatoes_id:
        raise PluginError('No criteria specified for rotten tomatoes lookup')

    def id_str() -> str:
        return f'<title={title},year={year},rottentomatoes_id={rottentomatoes_id}>'

    logger.debug('Looking up rotten tomatoes information for {}', id_str())

    movie = None

    # Try to lookup from cache
    if rottentomatoes_id:
        movie = (
            session.query(RottenTomatoesMovie)
            .filter(RottenTomatoesMovie.id == rottentomatoes_id)
            .first()
        )
    if not movie and title:
        movie_filter = session.query(RottenTomatoesMovie).filter(
            func.lower(RottenTomatoesMovie.title) == title.lower()
        )
        if year:
            movie_filter = movie_filter.filter(RottenTomatoesMovie.year == year)
        movie = movie_filter.first()
        if not movie:
            logger.debug('No matches in movie cache found, checking search cache.')
            found = (
                session.query(RottenTomatoesSearchResult)
                .filter(func.lower(RottenTomatoesSearchResult.search) == search_string)
                .first()
            )
            if found and found.movie:
                logger.debug('Movie found in search cache.')
                movie = found.movie
    if movie:
        # Movie found in cache, check if cache has expired.
        if movie.expired and not only_cached:
            logger.debug(
                'Cache has expired for {}, attempting to refresh from Rotten Tomatoes.', id_str()
            )
            try:
                result = movies_info(movie.id, api_key)
                movie = _set_movie_details(movie, session, result, api_key)
                session.merge(movie)
            except URLError:
                logger.error(
                    'Error refreshing movie details from Rotten Tomatoes, cached info being used.'
                )
        else:
            logger.debug('Movie {} information restored from cache.', id_str())
    else:
        if only_cached:
            raise PluginError('Movie %s not found from cache' % id_str())
        # There was no movie found in the cache, do a lookup from Rotten Tomatoes
        logger.debug('Movie {} not found in cache, looking up from rotten tomatoes.', id_str())
        try:
            if not movie and rottentomatoes_id:
                result = movies_info(rottentomatoes_id, api_key)
                if result:
                    movie = RottenTomatoesMovie()
                    movie = _set_movie_details(movie, session, result, api_key)
                    session.add(movie)

            if not movie and title:
                # TODO: Extract to method
                logger.verbose('Searching from rt `{}`', search_string)
                results = movies_search(search_string, api_key=api_key)
                if results:
                    results = results.get('movies')
                    if results:
                        for movie_res in results:
                            seq = difflib.SequenceMatcher(
                                lambda x: x == ' ', movie_res['title'].lower(), title.lower()
                            )
                            movie_res['match'] = seq.ratio()
                        results.sort(key=lambda x: x['match'], reverse=True)

                        # Remove all movies below MIN_MATCH, and different year
                        for movie_res in results[:]:

                            if year and movie_res.get('year'):
                                movie_res['year'] = int(movie_res['year'])
                                if movie_res['year'] != year:
                                    release_year = False
                                    if movie_res.get('release_dates', {}).get('theater'):
                                        logger.debug('Checking year against theater release date')
                                        release_year = time.strptime(
                                            movie_res['release_dates'].get('theater'), '%Y-%m-%d'
                                        ).tm_year
                                    elif movie_res.get('release_dates', {}).get('dvd'):
                                        logger.debug('Checking year against dvd release date')
                                        release_year = time.strptime(
                                            movie_res['release_dates'].get('dvd'), '%Y-%m-%d'
                                        ).tm_year
                                    if not (release_year and release_year == year):
                                        logger.debug(
                                            'removing {} - {} (wrong year: {})',
                                            movie_res['title'],
                                            movie_res['id'],
                                            str(release_year or movie_res['year']),
                                        )
                                        results.remove(movie_res)
                                        continue
                            if movie_res['match'] < MIN_MATCH:
                                logger.debug('removing {} (min_match)', movie_res['title'])
                                results.remove(movie_res)
                                continue

                        if not results:
                            raise PluginError('no appropiate results')

                        if len(results) == 1:
                            logger.debug('SUCCESS: only one movie remains')
                        else:
                            # Check min difference between best two hits
                            diff = results[0]['match'] - results[1]['match']
                            if diff < MIN_DIFF:
                                logger.debug(
                                    'unable to determine correct movie, min_diff too small(`{} ({}) - {}` <-?-> `{} ({}) - {}`)',
                                    results[0]['title'],
                                    results[0]['year'],
                                    results[0]['id'],
                                    results[1]['title'],
                                    results[1]['year'],
                                    results[1]['id'],
                                )
                                for r in results:
                                    logger.debug(
                                        'remain: {} (match: {}) {}',
                                        r['title'],
                                        r['match'],
                                        r['id'],
                                    )
                                raise PluginError('min_diff')

                        result = movies_info(results[0].get('id'), api_key)

                        if not result:
                            result = results[0]

                        movie = (
                            session.query(RottenTomatoesMovie)
                            .filter(RottenTomatoesMovie.id == result['id'])
                            .first()
                        )

                        if not movie:
                            movie = RottenTomatoesMovie()
                            movie = _set_movie_details(movie, session, result, api_key)
                            session.add(movie)
                            session.commit()

                        if title.lower() != movie.title.lower():
                            logger.debug("Saving search result for '{}'", search_string)
                            session.add(
                                RottenTomatoesSearchResult(search=search_string, movie=movie)
                            )
        except URLError:
            raise PluginError('Error looking up movie from RottenTomatoes')

    if not movie:
        raise PluginError('No results found from rotten tomatoes for %s' % id_str())
    else:
        # Access attributes to force the relationships to eager load before we detach from session
        for attr in [
            'alternate_ids',
            'cast',
            'directors',
            'genres',
            'links',
            'posters',
            'release_dates',
        ]:
            getattr(movie, attr)
        session.commit()
        return movie


# TODO: get rid of or heavily refactor
def _set_movie_details(
    movie: RottenTomatoesMovie,
    session: Session,
    movie_data: Optional[Dict[str, Any]] = None,
    api_key: Optional[str] = None,
) -> Any:
    """
    Populate ``movie`` object from given data

    :param movie: movie object to update
    :param session: session to use, returned Movie will be live in that session
    :param api_key: optionally specify an API key to use
    :param movie_data: data to copy into the :movie:
    """

    if not movie_data:
        if not movie.id:
            raise PluginError('Cannot get rotten tomatoes details without rotten tomatoes id')
        movie_data = movies_info(movie.id, api_key)
    if movie_data:
        if movie.id:
            logger.debug(
                "Updating movie info (actually just deleting the old info and adding the new)"
            )
            del movie.release_dates[:]
            del movie.posters[:]
            del movie.alternate_ids[:]
            del movie.links[:]
        movie.update_from_dict(movie_data)
        movie.update_from_dict(movie_data.get('ratings'))
        genres = movie_data.get('genres')
        if genres:
            for name in genres:
                genre = (
                    session.query(RottenTomatoesGenre)
                    .filter(func.lower(RottenTomatoesGenre.name) == name.lower())
                    .first()
                )
                if not genre:
                    genre = RottenTomatoesGenre(name)
                movie.genres.append(genre)
        release_dates = movie_data.get('release_dates')
        if release_dates:
            for name, date in list(release_dates.items()):
                movie.release_dates.append(ReleaseDate(name, date))
        posters = movie_data.get('posters')
        if posters:
            for name, url in list(posters.items()):
                movie.posters.append(RottenTomatoesPoster(name, url))
        cast = movie_data.get('abridged_cast')
        if cast:
            for res_actor in cast:
                actor = (
                    session.query(RottenTomatoesActor)
                    .filter(func.lower(RottenTomatoesActor.rt_id) == res_actor['id'])
                    .first()
                )
                if not actor:
                    actor = RottenTomatoesActor(res_actor['name'], res_actor['id'])
                movie.cast.append(actor)
        directors = movie_data.get('abridged_directors')
        if directors:
            for res_director in directors:
                director = (
                    session.query(RottenTomatoesDirector)
                    .filter(
                        func.lower(RottenTomatoesDirector.name) == res_director['name'].lower()
                    )
                    .first()
                )
                if not director:
                    director = RottenTomatoesDirector(res_director['name'])
                movie.directors.append(director)
        alternate_ids = movie_data.get('alternate_ids')
        if alternate_ids:
            for name, id in list(alternate_ids.items()):
                movie.alternate_ids.append(RottenTomatoesAlternateId(name, id))
        links = movie_data.get('links')
        if links:
            for name, url in list(links.items()):
                movie.links.append(RottenTomatoesLink(name, url))
        movie.updated = datetime.now()
    else:
        raise PluginError('No movie_data for rottentomatoes_id %s' % movie.id)

    return movie


def movies_info(id, api_key: Optional[str] = None):
    if not api_key:
        api_key = API_KEY
    url = f'{SERVER}/{API_VER}/movies/{id}.json?apikey={api_key}'
    result = get_json(url)
    if isinstance(result, dict) and result.get('id'):
        return result


def lists(
    list_type,
    list_name,
    limit: int = 20,
    page_limit: int = 20,
    page: Optional[int] = None,
    api_key=None,
):
    if isinstance(list_type, str):
        list_type = list_type.replace(' ', '_')
    if isinstance(list_name, str):
        list_name = list_name.replace(' ', '_')

    if not api_key:
        api_key = API_KEY

    url = f'{SERVER}/{API_VER}/lists/{list_type}/{list_name}.json?apikey={api_key}'
    if limit:
        url += f'&limit={limit}'
    if page_limit:
        url += f'&page_limit={page_limit}'
    if page:
        url += f'&page={page}'

    results = get_json(url)
    if isinstance(results, dict) and len(results.get('movies')):
        return results


def movies_search(
    q, page_limit: Optional[int] = None, page: Optional[int] = None, api_key: Optional[str] = None
):
    if isinstance(q, str):
        q = quote_plus(q.encode('latin-1', errors='ignore'))

    if not api_key:
        api_key = API_KEY

    url = f'{SERVER}/{API_VER}/movies.json?q={q}&apikey={api_key}'
    if page_limit:
        url += f'&page_limit={page_limit}'
    if page:
        url += f'&page={page}'

    results = get_json(url)
    if isinstance(results, dict) and results.get('total') and results.get('movies'):
        return results


def get_json(url: str) -> Optional[Dict[str, Any]]:
    try:
        logger.debug('fetching json at {}', url)
        data = session.get(url)
        return data.json()
    except requests.RequestException as e:
        logger.warning('Request failed {}: {}', url, e)
        return None
    except ValueError:
        logger.warning('Rotten Tomatoes returned invalid json at: {}', url)
        return None



# -*- coding: utf-8 -*-


import re
from urllib.request import ProxyHandler

import feedparser
from loguru import logger

from flexget import entry, plugin
from flexget.components.sites.utils import normalize_unicode
from flexget.entry import Entry
from flexget.event import event
from flexget.plugin import PluginError
from flexget.utils import qualities
from flexget.utils.requests import RequestException
from flexget.utils.soup import get_soup

try:
    from cloudscraper.exceptions import CaptchaException, CloudflareException

    cf_exceptions = (CloudflareException, CaptchaException)
except ModuleNotFoundError:
    cf_exceptions = ()
    pass  # cloudscraper module is optional

__authors__ = 'danfocus, Karlson2k'

logger = logger.bind(name='lostfilm')

RSS_TITLE_REGEXP = re.compile(
    r'^(?P<sr_rus>[^)(]+?)(?: \((?P<sr_org>[^)]+)\))?\. (?:(?P<ep_rus>.*)\. )?\(S(?P<season>\d+)E(?P<episode>\d+)\)$'
)
RSS_LINK_REGEXP = re.compile(
    r'^https?://[a-z./]+/series/(?P<sr_org2>.+)/season_(?P<season>\d+)/episode_(?P<episode>\d+)/$'
)
RSS_LF_ID_REGEXP = re.compile(r'/Images/(?P<id>\d+)/Posters/image\.(?:png|jpg|jpeg)')
PAGE_TEXT_REGEXP = re.compile(
    r'^\s*(?P<season>\d+)\s+сезон\s+(?P<episode>\d+)\s+серия\.(?:\s(?P<ep_rus>[^(]+?(?:\([^)]+?\))??))?(?:\s+\((?P<ep_org>[^)(]*?(?:\([^)]+?\))??)\s?\))?$'
)
PAGE_LINKMAIN_REGEXP = re.compile(
    r'(?:(?P<sr_rus>.+?)\.\s+)??(?:(?P<season>\d+) сезон, (?P<episode>\d+) серия\.\s+)?(?:(?P<ql>[0-9A-Za-z]+)\s)?(?P<tp>\b[A-Za-z-]*(?:Rip|RIP|rip))$'
)

quality_map = {
    'SD': '480p.mp3.xvid',
    '1080': '1080p.ac3.h264',
    'MP4': '720p.aac.h264',
    'HD': '720p.ac3.h264',
}

# All URLs must have '/' at the end
SITE_URLS = [
    'https://www.lostfilmtv.site/',
    'https://www.lostfilmtv.uno/',
    'https://www.lostfilm.run/',
    'https://www.lostfilm.uno/',
    'https://www.lostfilm.win/',
    'https://www.lostfilm.tv/',
]

SIMPLIFY_MAP = str.maketrans(
    {
        '&': ' and ',
        "'": None,
        '\\': None,
    }
)
SIMPLIFY_MAP.update(dict.fromkeys([ord(ch) for ch in '_./-,[](){}:;!?@#%^*+<>=~`$'], ' '))


class TextProcessingError(Exception):
    def __init__(self, value):
        super().__init__()
        self.value = value

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return repr(self.value)


class LostFilm:
    """
    Grab new torrents from lostfilm RSS feed

    Example:

      lostfilm: yes

    or

      lostfilm: <lf_session_cookie_value>

    Advanced usage:

      lostfilm:
        lf_session: <lf_session_cookie_value>
        prefilter: no
        site_urls:
          - "http://www.example.com/"
          - "https://www.example.org/"
    """

    schema = {
        'type': ['boolean', 'string', 'object'],
        'properties': {
            'lf_session': {'type': 'string'},
            'prefilter': {'type': 'boolean'},
            'site_urls': {
                'type': ['string', 'array'],
                'format': 'url',
                'items': {'type': 'string', 'format': 'url'},
            },
        },
        'additionalProperties': False,
    }

    def build_config(self, config):
        """Set defaults to config"""
        cfg = dict()
        if isinstance(config, bool):
            cfg['enabled'] = bool(config)
        elif isinstance(config, str):
            cfg['lf_session'] = str(config)
            cfg['enabled'] = True
        else:
            cfg = dict(config)
            cfg['enabled'] = True
        cfg.setdefault('prefilter', True)
        if isinstance(cfg.get('site_urls'), str):
            cfg['site_urls'] = [config['site_urls']]
        if not cfg.get('site_urls'):
            cfg['site_urls'] = SITE_URLS
        else:
            site_urls = []
            for url in cfg['site_urls']:
                if url.endswith('/'):
                    site_urls.append(url)
                else:
                    site_urls.append(url + "/")
            cfg['site_urls'] = site_urls
        return cfg

    def on_task_input(self, task, config):
        config = self.build_config(config)
        logger.trace('Config is {}', config)
        if not config['enabled']:
            return
        if config.get('lf_session') is not None:
            task.requests.cookies.set('lf_session', config['lf_session'])
            logger.debug('lf_session is set')
        task.requests.headers.update({'Cache-Control': 'no-cache', 'Pragma': 'no-cache'})
        prefilter_list = set()
        if config['prefilter']:
            prefilter_list = self._get_series(task)
            if prefilter_list:
                logger.verbose('Generated pre-filter list with {} entries', len(prefilter_list))
            else:
                logger.warning('Pre-filter list is empty. No series names are configured?')

        proxy_handler = None
        if task.requests.proxies is not None:
            proxy_handler = ProxyHandler(task.requests.proxies)

        site_urls = config['site_urls']
        tried_urls = []

        while site_urls:
            rss_url = site_urls[0] + "rss.xml"  # If RSS url changes, update it here
            logger.trace('Trying to get and parse the RSS feed: {}', rss_url)
            try:
                rss = feedparser.parse(
                    rss_url,
                    handlers=[proxy_handler],
                    request_headers={'Cache-Control': 'no-cache', 'Pragma': 'no-cache'},
                )
                status = rss.get('status')
                if status == 200:
                    logger.verbose('Received RSS feed from {}', rss_url)
                    break
                logger.info(
                    'Received {} status instead of 200 (OK) when trying to download the RSS feed {}',
                    status,
                    rss_url,
                )
            except Exception as e:
                logger.info('Cannot get or parse the RSS feed {}. Error: {}', rss_url, e)
            rss = None
            tried_urls.append(site_urls.pop(0))

        if not rss:
            raise PluginError('Cannot get the RSS feed')
        # Use failed site locations as the last resot option for the redirect page
        site_urls.extend(tried_urls)

        entries = []
        for idx, item in enumerate(rss.entries, 1):
            series_name_rus = series_name_org = None
            episode_name_rus = episode_name_org = None
            season_num = episode_num = None
            perfect_match = False

            if item.get('title') is None:
                logger.warning('RSS item doesn\'t have a title')
            else:
                logger.trace('Got RSS item title: {}', item['title'])
                title_match = RSS_TITLE_REGEXP.fullmatch(item['title'])
                if title_match is not None:
                    if title_match['sr_org'] is not None:
                        series_name_org = title_match['sr_org']
                        series_name_rus = title_match['sr_rus']
                        if title_match['ep_rus'] is not None:
                            perfect_match = True
                    else:
                        series_name_org = title_match['sr_rus']
                        series_name_rus = None
                    season_num = int(title_match['season'])
                    episode_num = int(title_match['episode'])
                    episode_name_rus = title_match['ep_rus']
                else:
                    logger.warning('Cannot parse RSS item title: {}', item['title'])

            # Skip series names that are not configured.
            # Do not filter out the current item if it is not matched perfectly.
            # It's better to process an extra item which will be filtered later by
            # series plugin than throw out actually needed item because incorrectly
            # matched name was not found in the pre-filter list.
            if prefilter_list:
                if perfect_match:
                    try:
                        folded_name = self._simplify_name(series_name_org)
                    except TextProcessingError as e:
                        logger.warning('RSS item series name "{}" could be wrong', series_name_org)
                        folded_name = None
                    if folded_name and folded_name not in prefilter_list:
                        if idx != len(rss.entries) or entries or task.no_entries_ok:
                            logger.debug(
                                'Skipping "{}" as "{}" was not found in the list of configured series',
                                item['title'],
                                series_name_org,
                            )
                            continue
                        else:
                            logger.debug(
                                'Force adding the last RSS item to the result to avoid warning of empty output'
                            )
                    else:
                        logger.trace(
                            '"{}" was found in the list of configured series', series_name_org
                        )
                else:
                    logger.debug(
                        'Not skipping RSS item as series names may be detected incorrectly'
                    )

            if item.get('description') is None:
                logger.warning('RSS item doesn\'t have a description, skipping')
                continue
            lostfilm_id_match = RSS_LF_ID_REGEXP.search(item['description'])
            if lostfilm_id_match is None or lostfilm_id_match['id'] is None:
                logger.warning(
                    'RSS item doesn\'t have lostfilm id in the description: {}, skipping'.item[
                        'description'
                    ]
                )
                continue
            lostfilm_id = int(lostfilm_id_match['id'])

            if not series_name_org or season_num is None or episode_num is None:
                if item.get('link') is None:
                    logger.warning('RSS item doesn\'t have a link, skipping')
                    continue
                link_match = RSS_LINK_REGEXP.fullmatch(item['link'])
                if link_match is None:
                    logger.warning('Cannot parse RSS item link, skipping: {}', item['link'])
                    continue
                series_name_org = link_match['sr_org2'].replace('_', ' ')
                season_num = int(link_match['season'])
                episode_num = int(link_match['episode'])
                logger.verbose('Using imprecise information from RSS item \'link\'')

            logger.trace(
                (
                    'Processing RSS entry: names: series "{}", series ru "{}", episode ru "{}"; '
                    'numbers: season "{}", episode "{}", lostfilm id "{}"; perfect detect: {}'
                ),
                series_name_org,
                series_name_rus,
                episode_name_rus,
                season_num,
                episode_num,
                lostfilm_id,
                perfect_match,
            )
            params = {'c': lostfilm_id, 's': season_num, 'e': episode_num}

            tried_urls = []
            while site_urls:
                redirect_url = site_urls[0] + 'v_search.php'
                logger.trace('Trying to get the redirect page: {}', redirect_url)
                try:
                    response = task.requests.get(redirect_url, params=params)
                    if response.status_code == 200:
                        logger.debug('The redirect page is downloaded from {}', redirect_url)
                        break
                    logger.verbose(
                        'Got status {} while retriving the redirect page {}',
                        response.status_code,
                        redirect_url,
                    )
                except RequestException as e:
                    logger.verbose(
                        'Failed to get the redirect page from {}. Error: {}', redirect_url, e
                    )
                except cf_exceptions as e:
                    logger.verbose(
                        'Cannot bypass CF page protection to get the redirect page {}. Error: {}',
                        redirect_url,
                        e,
                    )
                except Exception as e:
                    # Catch other errors related to download to avoid crash
                    logger.warning(
                        'Got unexpected exception when trying to get the redirect page. Error: {}',
                        redirect_url,
                        e,
                    )
                response = None
                tried_urls.append(site_urls.pop(0))

            # Use failed site locations as the last resot option for the next attempts
            site_urls.extend(tried_urls)

            if not response:
                if config.get('lf_session') is not None:
                    logger.error(
                        'Failed to get the redirect page. '
                        'Check whether "lf_session" parameter is correct.'
                    )
                else:
                    logger.error(
                        'Failed to get the redirect page. '
                        'Specify your "lf_session" cookie value in plugin parameters.'
                    )
                continue

            page = get_soup(response.content)

            download_page_url = None
            find_item = page.find('html', recursive=False)
            if find_item is not None:
                find_item = find_item.find('head', recursive=False)
                if find_item is not None:
                    find_item = find_item.find(
                        'meta', attrs={'http-equiv': "refresh"}, recursive=False
                    )
                    if (
                        find_item is not None
                        and find_item.has_attr('content')
                        and find_item['content'].startswith('0; url=http')
                    ):
                        download_page_url = find_item['content'][7:]
            if not download_page_url:
                if config.get('lf_session') is not None:
                    logger.error(
                        'Links were not foung on lostfilm.tv torrent download page. '
                        'Check whether "lf_session" parameter is correct.'
                    )
                else:
                    logger.error(
                        'Links were not foung on lostfilm.tv torrent download page. '
                        'Specify your "lf_session" cookie value in plugin parameters.'
                    )
                continue

            try:
                response = task.requests.get(download_page_url)
            except RequestException as e:
                logger.error('Failed to get the download page {}. Error: {}', download_page_url, e)
                continue
            except cf_exceptions as e:
                logger.error(
                    'Cannot pass CF page protection to get the download page {}. Error: {}',
                    download_page_url,
                    e,
                )
                continue
            except Exception as e:
                # Catch other errors related to download to avoid crash
                logger.error(
                    'Got unexpected exception when trying to get the download page {}. Error: {}',
                    download_page_url,
                    e,
                )
                continue

            page = get_soup(response.content)

            if not perfect_match:
                logger.trace('Trying to find series names in the final torrents download page')
                find_item = page.find('div', class_='inner-box--subtitle')
                if find_item is not None:
                    title_org_div = find_item.text.strip()
                    if title_org_div.endswith(', сериал') and len(title_org_div) != 8:
                        series_name_org = title_org_div[:-8]
                    else:
                        logger.info(
                            'Cannot parse text on the final download page for original series name'
                        )
                else:
                    logger.info('Cannot parse the final download page for original series name')

                find_item = page.find('div', class_='inner-box--title')
                if find_item is not None and find_item.text.strip():
                    series_name_rus = find_item.text.strip()
                else:
                    logger.info('Cannot parse the final download page for russian series name')

            find_item = page.find('div', class_='inner-box--text')
            if find_item is not None:
                info_match = PAGE_TEXT_REGEXP.fullmatch(find_item.text.strip())
                if info_match is not None:
                    if (
                        int(info_match['season']) != season_num
                        or int(info_match['episode']) != episode_num
                    ):
                        logger.warning(
                            (
                                'Using season number ({}) and episode number ({}) from download page instead of '
                                'season number ({}) and episode number ({}) in RSS item'
                            ),
                            int(info_match['season']),
                            int(info_match['episode']),
                            season_num,
                            episode_num,
                        )
                        season_num = int(info_match['season'])
                        eposode_num = int(info_match['episode'])
                    if info_match['ep_org'] is not None:
                        episode_name_org = info_match['ep_org'].strip()
                    if (
                        not perfect_match
                        and info_match['ep_rus'] is not None
                        and info_match['ep_rus'].strip()
                    ):
                        episode_name_rus = info_match['ep_rus'].strip()
                else:
                    logger.info('Cannot parse text on the final download page for episode names')
            else:
                logger.info('Cannot parse the final download page for episode names')

            r_type = ''
            find_item = page.find('div', class_='inner-box--link main')
            if find_item:
                find_item = find_item.find('a')
                if find_item:
                    info_match = PAGE_LINKMAIN_REGEXP.search(find_item.text)
                    if info_match:
                        r_type = info_match['tp']
                        logger.debug('Found rip type "{}"', r_type)

            if not series_name_org:
                find_item = item.get['title']
                if find_item:
                    logger.warning(
                        (
                            'Unable to detect series name. Full RSS item title will be used in hope '
                            'that series parser will be able to detect something: {}'
                        ),
                        find_item,
                    )
                    series_name_org = None
                else:
                    logger.error('Unable to detect series name. Skipping RSS item.')
                    continue

            d_items = page.find_all('div', class_='inner-box--item')
            if not d_items:
                logger.error('No download links were found on the download page')
                continue

            episode_id = 'S{:02d}E{:02d}'.format(season_num, episode_num)
            for d_item in d_items:
                find_item = d_item.find('div', class_='inner-box--link sub').a['href']
                if not find_item:
                    logger.warning('Download item does not have a link')
                    continue
                torrent_link = find_item

                find_item = d_item.find('div', class_='inner-box--label')
                if not find_item:
                    logger.warning('Download item does not have quality indicator')
                    continue
                lf_quality = find_item.text.strip()

                if quality_map.get(lf_quality):
                    quality = quality_map.get(lf_quality)
                else:
                    logger.info('Download item has unknown quality indicator: {}', lf_quality)
                    quality = lf_quality
                if series_name_org:
                    new_title = '.'.join(
                        [series_name_org, episode_id, quality, r_type, 'LostFilm.TV']
                    )
                else:
                    new_title = '{} {}'.format(item['title'], quality).strip()
                new_entry = Entry()
                new_entry['title'] = new_title
                new_entry['url'] = torrent_link
                if series_name_org:
                    new_entry['series_name'] = series_name_org
                    new_entry['series_name_org'] = series_name_org
                if perfect_match:
                    new_entry['series_exact'] = True
                new_entry['series_id'] = episode_id
                new_entry['series_id_type'] = 'ep'
                new_entry['series_season'] = season_num
                new_entry['series_episode'] = episode_num
                new_entry['series_episodes'] = 1
                new_entry['season_pack'] = None
                new_entry['proper'] = False
                new_entry['proper_count'] = 0
                new_entry['special'] = False
                new_entry['release_group'] = 'LostFilm.TV'
                if quality_map.get(lf_quality):
                    if r_type:
                        new_entry['quality'] = qualities.Quality('.'.join([quality, r_type]))
                    else:
                        new_entry['quality'] = qualities.Quality(quality)
                if series_name_rus:
                    new_entry['series_name_rus'] = series_name_rus
                if episode_name_rus:
                    new_entry['episode_name_rus'] = episode_name_rus
                if episode_name_org:
                    new_entry['episode_name_org'] = episode_name_org
                new_entry['lostfilm_id'] = lostfilm_id
                entries.append(new_entry)
                logger.trace(
                    (
                        'Added new entry: names: series "{}", series ru "{}", episode "{}", episode ru "{}"; '
                        'numbers: season "{}", episode "{}", lostfilm id "{}"; quality: "{}", perfect detect: {}'
                    ),
                    series_name_org,
                    series_name_rus,
                    episode_name_org,
                    episode_name_rus,
                    season_num,
                    episode_num,
                    lostfilm_id,
                    quality,
                    perfect_match,
                )

        return entries

    @staticmethod
    def _simplify_name(name: str) -> str:
        name = normalize_unicode(name)
        name = name.replace('&amp;', ' and ').translate(SIMPLIFY_MAP)
        name = ' '.join(name.split()).casefold()
        if not name:
            raise TextProcessingError('Simplified name of series is empty')
        return name

    @staticmethod
    def _get_series(task):
        if not task.config.get('series'):
            logger.warning('No series plugin in the task')
            return None

        names_list = set()
        try:
            # build list of all configured series names
            if isinstance(task.config['series'], list):
                # Flat list
                LostFilm._add_names_from_cfg_list(names_list, task.config['series'])
            elif isinstance(task.config['series'], dict):
                # Dictionary with groups
                for group, series_list in task.config['series'].items():
                    if group == 'settings':
                        continue
                    LostFilm._add_names_from_cfg_list(names_list, series_list)
            else:
                logger.warning('Unsupported series configuration type')
                return None
        except Exception as e:
            logger.warning('Error parsing series config: {:s}'.format(repr(e)))
            names_list = None

        return names_list

    @staticmethod
    def _add_names_from_cfg_list(names_list: set, cfg_list: list) -> None:
        for s_item in cfg_list:
            if isinstance(s_item, str):
                names_list.add(LostFilm._simplify_name(s_item))
            elif isinstance(s_item, int) or isinstance(s_item, float):
                # The name is something like '365' or '36.6'
                names_list.add(LostFilm._simplify_name(str(s_item)))
            elif isinstance(s_item, dict):
                s_name, s_cfg = list(s_item.items())[0]
                names_list.add(LostFilm._simplify_name(s_name))
                if s_cfg.get('alternate_name'):
                    if isinstance(s_cfg['alternate_name'], str):
                        names_list.add(LostFilm._simplify_name(s_cfg['alternate_name']))
                    elif isinstance(s_cfg['alternate_name'], list):
                        # A list of alternate names
                        for a_name in s_cfg['alternate_name']:
                            names_list.add(LostFilm._simplify_name(a_name))
                    else:
                        raise PluginError(
                            'Cannot read series "alternate_name" for "{:s}"'.format(s_name)
                        )
            else:
                raise PluginError(
                    'Series configuration list item has ' 'unsupported type: %s' % type(s_item)
                )


@event('plugin.register')
def register_plugin():
    plugin.register(LostFilm, 'lostfilm', api_ver=2)



# -*- coding: utf-8 -*-
from typing import Optional

from loguru import logger

from flexget import entry, plugin
from flexget.event import event
from flexget.manager import Session

from . import api_trakt as plugin_api_trakt
from . import db

# TODO: not very nice ..
lookup_series = plugin_api_trakt.ApiTrakt.lookup_series
lookup_movie = plugin_api_trakt.ApiTrakt.lookup_movie


logger = logger.bind(name='trakt_lookup')


def is_show(entry: entry.Entry) -> bool:
    return entry.get('series_name') or entry.get('tvdb_id', eval_lazy=False)


def is_episode(entry: entry.Entry) -> bool:
    return entry.get('series_season') and entry.get('series_episode')


def is_season(entry: entry.Entry) -> bool:
    return entry.get('series_season') and not is_episode(entry)


def is_movie(entry: entry.Entry) -> bool:
    return bool(entry.get('movie_name'))


def get_media_type_for_entry(entry: entry.Entry) -> Optional[str]:
    if is_episode(entry):
        return 'episode'
    elif is_season(entry):
        return 'season'
    elif is_show(entry):
        return 'show'
    elif is_movie(entry):
        return 'movie'
    return None


@entry.register_lazy_lookup('trakt_user_data_lookup')
def trakt_user_data_lookup(entry, field_name, data_type, media_type, username, account):
    trakt = plugin_api_trakt.ApiTrakt(username=username, account=account)
    user_data_lookup = trakt.lookup_map[data_type][media_type]

    try:
        with Session() as session:
            result = user_data_lookup(get_db_data_for(media_type, entry, session), entry['title'])
    except LookupError as e:
        logger.debug(e)
    else:
        entry[field_name] = result

    return entry


def _get_lookup_args(entry: entry.Entry) -> dict:
    args = {
        'title': entry.get('series_name', eval_lazy=False) or entry.get('title', eval_lazy=False),
        'year': entry.get('year', eval_lazy=False) or entry.get('movie_year', eval_lazy=False),
        'trakt_slug': (
            entry.get('trakt_show_slug', eval_lazy=False)
            or entry.get('trakt_movie_slug', eval_lazy=False)
        ),
        'tmdb_id': entry.get('tmdb_id', eval_lazy=False),
        'tvdb_id': entry.get('tvdb_id', eval_lazy=False),
        'imdb_id': entry.get('imdb_id', eval_lazy=False),
        'tvrage_id': entry.get('tvrage_id', eval_lazy=False),
    }

    if entry.get('trakt_movie_id', eval_lazy=False):
        args['trakt_id'] = entry['trakt_movie_id']
    elif entry.get('trakt_show_id', eval_lazy=False):
        args['trakt_id'] = entry['trakt_show_id']
    elif is_movie(entry) and entry.get('trakt_movie_id', eval_lazy=True):
        args['trakt_id'] = entry['trakt_movie_id']
    elif entry.get('trakt_show_id', eval_lazy=True):
        args['trakt_id'] = entry['trakt_show_id']

    return args


def get_db_data_for(data_type: str, entry: entry.Entry, session: Session):
    if data_type == 'movie':
        movie_lookup_args = _get_lookup_args(entry)
        return lookup_movie(session=session, **movie_lookup_args)
    series_lookup_args = _get_lookup_args(entry)
    show = lookup_series(session=session, **series_lookup_args)
    if data_type == 'show':
        return show
    if data_type == 'season':
        return show.get_season(entry['series_season'], session)
    if data_type == 'episode':
        return show.get_episode(entry['series_season'], entry['series_episode'], session)


lazy_lookup_types = {
    'show': {
        'trakt_series_name': 'title',
        'trakt_series_year': 'year',
        'imdb_id': 'imdb_id',
        'tvdb_id': 'tvdb_id',
        'tmdb_id': 'tmdb_id',
        'trakt_show_id': 'id',
        'trakt_show_slug': 'slug',
        'tvrage_id': 'tvrage_id',
        'trakt_trailer': 'trailer',
        'trakt_homepage': 'homepage',
        'trakt_series_runtime': 'runtime',
        'trakt_series_first_aired': 'first_aired',
        'trakt_series_air_time': 'air_time',
        'trakt_series_air_day': 'air_day',
        'trakt_series_content_rating': 'certification',
        'trakt_genres': lambda i: [db_genre.name for db_genre in i.genres],
        'trakt_series_network': 'network',
        'imdb_url': lambda series: series.imdb_id
        and 'http://www.imdb.com/title/%s' % series.imdb_id,
        'trakt_series_url': lambda series: series.slug
        and 'https://trakt.tv/shows/%s' % series.slug,
        'trakt_series_country': 'country',
        'trakt_series_status': 'status',
        'trakt_series_overview': 'overview',
        'trakt_series_rating': 'rating',
        'trakt_series_votes': 'votes',
        'trakt_series_language': 'language',
        'trakt_series_aired_episodes': 'aired_episodes',
        'trakt_series_episodes': lambda show: [episodes.title for episodes in show.episodes],
        'trakt_languages': 'translation_languages',
    },
    'show_actors': {'trakt_actors': lambda show: db.list_actors(show.actors)},
    'show_translations': {
        'trakt_translations': lambda show: db.get_translations_dict(show.translations, 'show')
    },
    'season': {
        'trakt_season_name': 'title',
        'trakt_season_tvdb_id': 'tvdb_id',
        'trakt_season_tmdb_id': 'tmdb_id',
        'trakt_season_tvrage': 'tvrage_id',
        'trakt_season_id': 'id',
        'trakt_season_first_aired': 'first_aired',
        'trakt_season_overview': 'overview',
        'trakt_season_episode_count': 'episode_count',
        'trakt_season': 'number',
        'trakt_season_aired_episodes': 'aired_episodes',
    },
    'episode': {
        'trakt_ep_name': 'title',
        'trakt_ep_imdb_id': 'imdb_id',
        'trakt_ep_tvdb_id': 'tvdb_id',
        'trakt_ep_tmdb_id': 'tmdb_id',
        'trakt_ep_tvrage': 'tvrage_id',
        'trakt_episode_id': 'id',
        'trakt_ep_first_aired': 'first_aired',
        'trakt_ep_overview': 'overview',
        'trakt_ep_abs_number': 'number_abs',
        'trakt_season': 'season',
        'trakt_episode': 'number',
        'trakt_ep_id': lambda ep: 'S%02dE%02d' % (ep.season, ep.number),
    },
    'movie': {
        'movie_name': 'title',
        'movie_year': 'year',
        'trakt_movie_name': 'title',
        'trakt_movie_year': 'year',
        'trakt_movie_id': 'id',
        'trakt_movie_slug': 'slug',
        'imdb_id': 'imdb_id',
        'tmdb_id': 'tmdb_id',
        'trakt_tagline': 'tagline',
        'trakt_overview': 'overview',
        'trakt_released': 'released',
        'trakt_runtime': 'runtime',
        'trakt_rating': 'rating',
        'trakt_votes': 'votes',
        'trakt_homepage': 'homepage',
        'trakt_trailer': 'trailer',
        'trakt_language': 'language',
        'trakt_genres': lambda i: [db_genre.name for db_genre in i.genres],
        'trakt_languages': 'translation_languages',
    },
    'movie_actors': {'trakt_actors': lambda movie: db.list_actors(movie.actors)},
    'movie_translations': {
        'trakt_translations': lambda movie: db.get_translations_dict(movie.translations, 'movie')
    },
}


@entry.register_lazy_lookup('trakt_lazy_lookup')
def lazy_lookup(entry, lazy_lookup_name, media_type):
    with Session() as session:
        try:
            db_data = get_db_data_for(media_type, entry, session)
        except LookupError as e:
            logger.debug(e)
        else:
            entry.update_using_map(lazy_lookup_types[lazy_lookup_name], db_data)
    return entry


def add_lazy_fields(entry: entry.Entry, lazy_lookup_name: str, media_type: str) -> None:
    """
    Adds lazy fields for one of the lookups in our `lazy_lookup_types` dict.

    :param entry: The entry to add lazy fields to.
    :param lazy_lookup_name: One of the keys in `lazy_lookup_types` dict.
    :param media_type: show/season/episode/movie (the type of db data needed for this lazy lookup)
    """
    entry.add_lazy_fields(
        lazy_lookup, lazy_lookup_types[lazy_lookup_name], args=(lazy_lookup_name, media_type)
    )


user_data_fields = {
    'collected': 'trakt_collected',
    'watched': 'trakt_watched',
    'ratings': {
        'show': 'trakt_series_user_rating',
        'season': 'trakt_season_user_rating',
        'episode': 'trakt_ep_user_rating',
        'movie': 'trakt_movie_user_rating',
    },
}


def add_lazy_user_fields(
    entry: entry.Entry, data_type: str, media_type: str, username: str, account: str
) -> None:
    """
    Adds one of the user field lazy lookups to an entry.

    :param entry: Entry to add lazy fields to
    :param data_type: ratings/collected/watched (one of the keys in `user_data_fields` dict.)
    :param media_type: show/season/episode/movie
    :param username: Either this or account is required, the other can be None
    :param account: Either this or username is required, the other can be None
    """
    field_name = user_data_fields[data_type]
    if data_type == 'ratings':
        field_name = field_name[media_type]
    entry.add_lazy_fields(
        trakt_user_data_lookup,
        [field_name],
        args=(field_name, data_type, media_type, username, account),
    )


class PluginTraktLookup:
    """Retrieves trakt information for entries. Uses series_name,
    series_season, series_episode from series plugin.

    Example:

    trakt_lookup: yes

    Primarily used for passing trakt information to other plugins.
    Among these is the IMDB url for the series.

    This information is provided (via entry):
    series info:
    trakt_series_name
    trakt_series_runtime
    trakt_series_first_aired_epoch
    trakt_series_first_aired_iso
    trakt_series_air_time
    trakt_series_content_ratingi
    trakt_series_genres
    trakt_series_imdb_url
    trakt_series_trakt_url
    imdb_id
    tvdb_id
    trakt_series_actors
    trakt_series_country
    trakt_series_year
    trakt_series_tvrage_id
    trakt_series_status
    trakt_series_overview

    trakt_ep_name
    trakt_ep_season
    trakt_ep_number
    trakt_ep_overview
    trakt_ep_first_aired_epoch
    trakt_ep_first_aired_iso
    trakt_ep_id
    trakt_ep_tvdb_id


    """

    schema = {
        'oneOf': [
            {
                'type': 'object',
                'properties': {'account': {'type': 'string'}, 'username': {'type': 'string'}},
                'anyOf': [{'required': ['username']}, {'required': ['account']}],
                'error_anyOf': 'At least one of `username` or `account` options are needed.',
                'additionalProperties': False,
            },
            {'type': 'boolean'},
        ]
    }

    def on_task_start(self, task, config):
        if not isinstance(config, dict):
            config = {}

        self.trakt = plugin_api_trakt.ApiTrakt(
            username=config.get('username'), account=config.get('account')
        )

    # Run after series and metainfo series
    @plugin.priority(110)
    def on_task_metainfo(self, task, config):
        if not config:
            return

        if isinstance(config, bool):
            config = dict()

        for entry in task.entries:
            if is_show(entry):
                add_lazy_fields(entry, 'show', media_type='show')
                add_lazy_fields(entry, 'show_actors', media_type='show')
                add_lazy_fields(entry, 'show_translations', media_type='show')
                if is_episode(entry):
                    add_lazy_fields(entry, 'episode', media_type='episode')
                elif is_season(entry):
                    add_lazy_fields(entry, 'season', media_type='season')
            else:
                add_lazy_fields(entry, 'movie', media_type='movie')
                add_lazy_fields(entry, 'movie_actors', media_type='movie')
                add_lazy_fields(entry, 'movie_translations', media_type='movie')

            if config.get('username') or config.get('account'):
                credentials = {
                    'username': config.get('username'),
                    'account': config.get('account'),
                }
                media_type = get_media_type_for_entry(entry)
                if media_type:
                    add_lazy_user_fields(entry, 'collected', media_type=media_type, **credentials)
                    add_lazy_user_fields(entry, 'watched', media_type=media_type, **credentials)
                if is_show(entry):
                    add_lazy_user_fields(entry, 'ratings', media_type='show', **credentials)
                    if is_season(entry):
                        add_lazy_user_fields(entry, 'ratings', media_type='season', **credentials)
                    if is_episode(entry):
                        add_lazy_user_fields(entry, 'ratings', media_type='episode', **credentials)
                elif is_movie(entry):
                    add_lazy_user_fields(entry, 'ratings', media_type='movie', **credentials)

    @property
    def series_identifier(self):
        """Returns the plugin main identifier type"""
        return 'trakt_show_id'

    @property
    def movie_identifier(self):
        """Returns the plugin main identifier type"""
        return 'trakt_movie_id'


@event('plugin.register')
def register_plugin():
    plugin.register(
        PluginTraktLookup,
        'trakt_lookup',
        api_ver=2,
        interfaces=['task', 'series_metainfo', 'movie_metainfo'],
    )



import copy
from math import ceil

from flask import jsonify, request
from flask_restx import inputs
from sqlalchemy.orm.exc import NoResultFound

from flexget import plugin
from flexget.api import APIClient, APIResource, api
from flexget.api.app import (
    BadRequest,
    Conflict,
    NotFoundError,
    base_message_schema,
    etag,
    pagination_headers,
    success_response,
)
from flexget.event import fire_event
from flexget.plugin import PluginError

from . import db
from .utils import normalize_series_name

try:
    # NOTE: Importing other plugins is discouraged!
    from flexget.components.thetvdb.api import ObjectsContainer as tvdb
except ImportError:
    raise plugin.DependencyError(issued_by=__name__, missing='tvdb_lookup')


try:
    # NOTE: Importing other plugins is discouraged!
    from flexget.components.tvmaze.api import ObjectsContainer as tvmaze
except ImportError:
    raise plugin.DependencyError(issued_by=__name__, missing='tvmaze_lookup')

series_api = api.namespace('series', description='FlexGet Series operations')


def series_details(show, begin=False, latest=False):
    series_dict = {
        'id': show.id,
        'name': show.name,
        'alternate_names': [n.alt_name for n in show.alternate_names],
        'in_tasks': [_show.name for _show in show.in_tasks],
    }
    if begin:
        series_dict['begin_episode'] = show.begin.to_dict() if show.begin else None
    if latest:
        latest_entity = db.get_latest_release(show)
        series_dict['latest_entity'] = latest_entity.to_dict() if latest_entity else None
        if latest_entity:
            series_dict['latest_entity']['latest_release'] = latest_entity.latest_release.to_dict()
    return series_dict


class ObjectsContainer:
    episode_release_object = {
        'type': 'object',
        'properties': {
            'id': {'type': 'integer'},
            'title': {'type': 'string'},
            'downloaded': {'type': 'boolean'},
            'quality': {'type': 'string'},
            'proper_count': {'type': 'integer'},
            'first_seen': {'type': 'string', 'format': 'date-time'},
            'episode_id': {'type': 'integer'},
        },
        'required': [
            'id',
            'title',
            'downloaded',
            'quality',
            'proper_count',
            'first_seen',
            'episode_id',
        ],
    }

    season_release_object = copy.deepcopy(episode_release_object)
    del season_release_object['properties']['episode_id']
    season_release_object['properties']['season_id'] = {'type': 'integer'}
    season_release_object['required'].remove('episode_id')
    season_release_object['required'].append('season_id')

    episode_release_list_schema = {'type': 'array', 'items': episode_release_object}
    season_release_list_schema = {'type': 'array', 'items': season_release_object}

    episode_object = {
        'type': ['object', 'null'],
        'properties': {
            "first_seen": {'type': ['string', 'null'], 'format': 'date-time'},
            "id": {'type': 'integer'},
            "identified_by": {'type': 'string'},
            "identifier": {'type': 'string'},
            "premiere": {'type': ['string', 'boolean']},
            "number": {'type': 'integer'},
            "season": {'type': 'integer'},
            "series_id": {'type': 'integer'},
            "number_of_releases": {'type': 'integer'},
            'latest_release': episode_release_object,
        },
        'required': [
            'first_seen',
            'id',
            'identified_by',
            'identifier',
            'premiere',
            'number',
            'season',
            'series_id',
            'number_of_releases',
        ],
    }

    season_object = copy.deepcopy(episode_object)
    del season_object['properties']['number']
    season_object['required'].remove('number')
    season_object['required'].remove('premiere')

    single_series_object = {
        'type': 'object',
        'properties': {
            'id': {'type': 'integer'},
            'name': {'type': 'string'},
            'alternate_names': {'type': 'array', 'items': {'type': 'string'}},
            'in_tasks': {'type': 'array', 'items': {'type': 'string'}},
            'lookup': {
                'type': 'object',
                'properties': {
                    'tvmaze': tvmaze.tvmaze_series_object,
                    'tvdb': tvdb.tvdb_series_object,
                },
            },
            'latest_episode': episode_object,
            'begin_episode': episode_object,
        },
        'required': ['id', 'name', 'alternate_names', 'in_tasks'],
    }

    series_list_schema = {'type': 'array', 'items': single_series_object}

    episode_list_schema = {'type': 'array', 'items': episode_object}

    seasons_list_schema = {'type': 'array', 'items': season_object}

    series_edit_object = {
        'type': 'object',
        'properties': {
            'begin_episode': {'type': ['string', 'integer'], 'format': 'episode_or_season_id'},
            'alternate_names': {'type': 'array', 'items': {'type': 'string'}},
        },
        'anyOf': [{'required': ['begin_episode']}, {'required': ['alternate_names']}],
        'additionalProperties:': False,
    }

    series_input_object = copy.deepcopy(series_edit_object)
    series_input_object['properties']['name'] = {'type': 'string'}
    del series_input_object['anyOf']
    series_input_object['required'] = ['name']


series_list_schema = api.schema_model('list_series', ObjectsContainer.series_list_schema)
series_edit_schema = api.schema_model('series_edit_schema', ObjectsContainer.series_edit_object)
series_input_schema = api.schema_model('series_input_schema', ObjectsContainer.series_input_object)
show_details_schema = api.schema_model('show_details', ObjectsContainer.single_series_object)

episode_list_schema = api.schema_model('episode_list', ObjectsContainer.episode_list_schema)
episode_schema = api.schema_model('episode_item', ObjectsContainer.episode_object)

season_list_schema = api.schema_model('season_list', ObjectsContainer.seasons_list_schema)
season_schema = api.schema_model('episode_item', ObjectsContainer.season_object)

episode_release_schema = api.schema_model(
    'release_schema', ObjectsContainer.episode_release_object
)
episode_release_list_schema = api.schema_model(
    'release_list_schema', ObjectsContainer.episode_release_list_schema
)

season_release_schema = api.schema_model('release_schema', ObjectsContainer.season_release_object)
season_release_list_schema = api.schema_model(
    'release_list_schema', ObjectsContainer.season_release_list_schema
)

base_series_parser = api.parser()
base_series_parser.add_argument(
    'begin', type=inputs.boolean, default=True, help='Show series begin episode'
)
base_series_parser.add_argument(
    'latest',
    type=inputs.boolean,
    default=True,
    help='Show series latest downloaded episode and release',
)

sort_choices = ('show_name', 'last_download_date')
series_list_parser = api.pagination_parser(base_series_parser, sort_choices=sort_choices)
series_list_parser.add_argument(
    'in_config',
    choices=('configured', 'unconfigured', 'all'),
    default='configured',
    help="Filter list if shows are currently in configuration.",
)
series_list_parser.add_argument(
    'premieres', type=inputs.boolean, default=False, help="Filter by downloaded premieres only."
)
series_list_parser.add_argument(
    'lookup',
    choices=('tvdb', 'tvmaze'),
    action='append',
    help="Get lookup result for every show by sending another request to lookup API",
)
series_list_parser.add_argument('query', help="Search by name based on the query")

ep_identifier_doc = (
    "'episode_identifier' should be one of SxxExx, integer or date formatted such as 2012-12-12"
)


@series_api.route('/')
class SeriesAPI(APIResource):
    @etag
    @api.response(200, 'Series list retrieved successfully', series_list_schema)
    @api.response(NotFoundError)
    @api.doc(parser=series_list_parser, description="Get a  list of Flexget's shows in DB")
    def get(self, session=None):
        """List existing shows"""
        args = series_list_parser.parse_args()

        # Filter params
        configured = args['in_config']
        premieres = args['premieres']

        # Pagination and sorting params
        page = args['page']
        per_page = args['per_page']
        sort_by = args['sort_by']
        sort_order = args['order']
        name = normalize_series_name(args['query']) if args['query'] else None

        # Handle max size limit
        if per_page > 100:
            per_page = 100

        descending = sort_order == 'desc'

        # Data params
        lookup = args.get('lookup')
        begin = args.get('begin')
        latest = args.get('latest')

        start = per_page * (page - 1)
        stop = start + per_page

        kwargs = {
            'configured': configured,
            'premieres': premieres,
            'start': start,
            'stop': stop,
            'sort_by': sort_by,
            'descending': descending,
            'session': session,
            'name': name,
        }

        total_items = db.get_series_summary(count=True, **kwargs)

        if not total_items:
            return jsonify([])

        series_list = []
        for s in db.get_series_summary(**kwargs):
            series_object = series_details(s, begin, latest)
            series_list.append(series_object)

        # Total number of pages
        total_pages = int(ceil(total_items / float(per_page)))

        if total_pages < page and total_pages != 0:
            raise NotFoundError('page %s does not exist' % page)

        # Actual results in page
        actual_size = min(per_page, len(series_list))

        # Do relevant lookups
        if lookup:
            api_client = APIClient()
            for endpoint in lookup:
                base_url = '/%s/series/' % endpoint
                for show in series_list:
                    pos = series_list.index(show)
                    series_list[pos].setdefault('lookup', {})
                    url = base_url + show['name'] + '/'
                    result = api_client.get_endpoint(url)
                    series_list[pos]['lookup'].update({endpoint: result})

        # Get pagination headers
        pagination = pagination_headers(total_pages, total_items, actual_size, request)

        # Created response
        rsp = jsonify(series_list)

        # Add link header to response
        rsp.headers.extend(pagination)
        return rsp

    @api.response(201, model=show_details_schema)
    @api.response(Conflict)
    @api.validate(series_input_schema, description=ep_identifier_doc)
    def post(self, session):
        """Create a new show and set its first accepted episode and/or alternate names"""
        data = request.json
        series_name = data.get('name')

        normalized_name = normalize_series_name(series_name)
        matches = db.shows_by_exact_name(normalized_name, session=session)
        if matches:
            raise Conflict('Show `%s` already exist in DB' % series_name)
        show = db.Series()
        show.name = series_name
        session.add(show)

        ep_id = data.get('begin_episode')
        alt_names = data.get('alternate_names')
        if ep_id:
            db.set_series_begin(show, ep_id)
        if alt_names:
            try:
                db.set_alt_names(alt_names, show, session)
            except PluginError as e:
                # Alternate name already exist for a different show
                raise Conflict(e.value)
        session.commit()
        rsp = jsonify(series_details(show, begin=ep_id is not None))
        rsp.status_code = 201
        return rsp


@series_api.route('/search/<string:name>/')
@api.doc(
    description='Searches for a show in the DB via its name. Returns a list of matching shows.'
)
class SeriesGetShowsAPI(APIResource):
    @etag
    @api.response(200, 'Show list retrieved successfully', series_list_schema)
    @api.doc(params={'name': 'Name of the show(s) to search'}, parser=base_series_parser)
    def get(self, name, session):
        """List of shows matching lookup name"""
        name = normalize_series_name(name)
        matches = db.shows_by_name(name, session=session)

        args = series_list_parser.parse_args()
        begin = args.get('begin')
        latest = args.get('latest')

        shows = []
        for match in matches:
            shows.append(series_details(match, begin, latest))

        return jsonify(shows)


delete_parser = api.parser()
delete_parser.add_argument(
    'forget',
    type=inputs.boolean,
    default=False,
    help="Enabling this will fire a 'forget' event that will delete the downloaded releases "
    "from the entire DB, enabling to re-download them",
)


@series_api.route('/<int:show_id>/')
@api.doc(params={'show_id': 'ID of the show'})
@api.response(NotFoundError)
class SeriesShowAPI(APIResource):
    @etag
    @api.response(200, 'Show information retrieved successfully', show_details_schema)
    @api.doc(description='Get a specific show using its ID', parser=base_series_parser)
    def get(self, show_id, session):
        """Get show details by ID"""
        try:
            show = db.show_by_id(show_id, session=session)
        except NoResultFound:
            raise NotFoundError('Show with ID %s not found' % show_id)

        args = series_list_parser.parse_args()
        begin = args.get('begin')
        latest = args.get('latest')

        return jsonify(series_details(show, begin, latest))

    @api.response(200, 'Removed series from DB', model=base_message_schema)
    @api.doc(description='Delete a specific show using its ID', parser=delete_parser)
    def delete(self, show_id, session):
        """Remove series from DB"""
        try:
            show = db.show_by_id(show_id, session=session)
        except NoResultFound:
            raise NotFoundError('Show with ID %s not found' % show_id)

        name = show.name
        args = delete_parser.parse_args()
        db.remove_series(name, forget=args.get('forget'))

        return success_response('successfully removed series %s from DB' % show_id)

    @api.response(
        200, 'Episodes for series will be accepted starting with ep_id', show_details_schema
    )
    @api.response(Conflict)
    @api.validate(series_edit_schema, description=ep_identifier_doc)
    @api.doc(
        description='Set a begin episode or alternate names using a show ID. Note that alternate names override '
        'the existing names (if name does not belong to a different show).'
    )
    def put(self, show_id, session):
        """Set the initial episode of an existing show"""
        try:
            show = db.show_by_id(show_id, session=session)
        except NoResultFound:
            raise NotFoundError('Show with ID %s not found' % show_id)
        data = request.json
        ep_id = data.get('begin_episode')
        alt_names = data.get('alternate_names')
        if ep_id:
            db.set_series_begin(show, ep_id)
        if alt_names:
            try:
                db.set_alt_names(alt_names, show, session)
            except PluginError as e:
                # Alternate name already exist for a different show
                raise Conflict(e.value)
        return jsonify(series_details(show, begin=ep_id is not None))


entity_parser = api.pagination_parser(add_sort=True)


@api.response(NotFoundError)
@series_api.route('/<int:show_id>/seasons/')
@api.doc(
    params={'show_id': 'ID of the show'},
    description='The \'Series-ID\' header will be appended to the result headers',
)
class SeriesSeasonsAPI(APIResource):
    @etag
    @api.response(200, 'Seasons retrieved successfully for show', season_list_schema)
    @api.doc(description='Get all show seasons via its ID', parser=entity_parser)
    def get(self, show_id, session):
        """Get seasons by show ID"""
        args = entity_parser.parse_args()

        # Pagination and sorting params
        page = args['page']
        per_page = args['per_page']
        sort_order = args['order']

        # Handle max size limit
        if per_page > 100:
            per_page = 100

        descending = sort_order == 'desc'

        start = per_page * (page - 1)
        stop = start + per_page

        kwargs = {'start': start, 'stop': stop, 'descending': descending, 'session': session}

        try:
            show = db.show_by_id(show_id, session=session)
        except NoResultFound:
            raise NotFoundError('show with ID %s not found' % show_id)

        total_items = db.show_seasons(show, count=True, session=session)

        if not total_items:
            return jsonify([])

        seasons = [season.to_dict() for season in db.show_seasons(show, **kwargs)]

        total_pages = int(ceil(total_items / float(per_page)))

        if total_pages < page and total_pages != 0:
            raise NotFoundError('page %s does not exist' % page)

        # Actual results in page
        actual_size = min(per_page, len(seasons))

        # Get pagination headers
        pagination = pagination_headers(total_pages, total_items, actual_size, request)

        # Created response
        rsp = jsonify(seasons)

        # Add link header to response
        rsp.headers.extend(pagination)

        # Add series ID header
        rsp.headers.extend({'Series-ID': show_id})
        return rsp

    @api.response(200, 'Successfully forgotten all seasons from show', model=base_message_schema)
    @api.doc(
        description='Delete all show seasons via its ID. Deleting a season will mark it as wanted again',
        parser=delete_parser,
    )
    def delete(self, show_id, session):
        """Deletes all seasons of a show"""
        try:
            show = db.show_by_id(show_id, session=session)
        except NoResultFound:
            raise NotFoundError('show with ID %s not found' % show_id)
        args = delete_parser.parse_args()
        forget = args.get('forget')
        for season in show.seasons:
            db.remove_series_entity(show.name, season.identifier, forget)
        return success_response('successfully removed all series %s seasons from DB' % show_id)


@api.response(NotFoundError)
@api.response(BadRequest)
@series_api.route('/<int:show_id>/seasons/<int:season_id>/')
@api.doc(params={'show_id': 'ID of the show', 'season_id': 'Season ID'})
class SeriesSeasonsAPI(APIResource):
    @etag
    @api.response(200, 'Season retrieved successfully for show', season_schema)
    @api.doc(description='Get a specific season via its ID and show ID')
    def get(self, show_id, season_id, session):
        """Get season by show ID and season ID"""
        try:
            db.show_by_id(show_id, session=session)
        except NoResultFound:
            raise NotFoundError('show with ID %s not found' % show_id)
        try:
            season = db.season_by_id(season_id, session)
        except NoResultFound:
            raise NotFoundError('season with ID %s not found' % season_id)
        if not db.season_in_show(show_id, season_id):
            raise BadRequest('season with id %s does not belong to show %s' % (season_id, show_id))

        rsp = jsonify(season.to_dict())

        # Add Series-ID header
        rsp.headers.extend({'Series-ID': show_id})
        return rsp

    @api.response(200, 'Season successfully forgotten for show', model=base_message_schema)
    @api.doc(
        description='Delete a specific season via its ID and show ID. Deleting a season will mark it as '
        'wanted again',
        parser=delete_parser,
    )
    def delete(self, show_id, season_id, session):
        """Forgets season by show ID and season ID"""
        try:
            show = db.show_by_id(show_id, session=session)
        except NoResultFound:
            raise NotFoundError('show with ID %s not found' % show_id)
        try:
            season = db.season_by_id(season_id, session)
        except NoResultFound:
            raise NotFoundError('season with ID %s not found' % season_id)
        if not db.season_in_show(show_id, season_id):
            raise BadRequest('season with id %s does not belong to show %s' % (season_id, show_id))

        args = delete_parser.parse_args()
        db.remove_series_entity(show.name, season.identifier, args.get('forget'))

        return success_response(
            'successfully removed season %s from show %s' % (season_id, show_id)
        )


@api.response(NotFoundError)
@series_api.route('/<int:show_id>/episodes/')
@api.doc(
    params={'show_id': 'ID of the show'},
    description='The \'Series-ID\' header will be appended to the result headers',
)
class SeriesEpisodesAPI(APIResource):
    @etag
    @api.response(200, 'Episodes retrieved successfully for show', episode_list_schema)
    @api.doc(description='Get all show episodes via its ID', parser=entity_parser)
    def get(self, show_id, session):
        """Get episodes by show ID"""
        args = entity_parser.parse_args()

        # Pagination and sorting params
        page = args['page']
        per_page = args['per_page']
        sort_order = args['order']

        # Handle max size limit
        if per_page > 100:
            per_page = 100

        descending = sort_order == 'desc'

        start = per_page * (page - 1)
        stop = start + per_page

        kwargs = {'start': start, 'stop': stop, 'descending': descending, 'session': session}

        try:
            show = db.show_by_id(show_id, session=session)
        except NoResultFound:
            raise NotFoundError('show with ID %s not found' % show_id)

        total_items = db.show_episodes(show, count=True, session=session)

        if not total_items:
            return jsonify([])

        episodes = [episode.to_dict() for episode in db.show_episodes(show, **kwargs)]

        total_pages = int(ceil(total_items / float(per_page)))

        if total_pages < page and total_pages != 0:
            raise NotFoundError('page %s does not exist' % page)

        # Actual results in page
        actual_size = min(per_page, len(episodes))

        # Get pagination headers
        pagination = pagination_headers(total_pages, total_items, actual_size, request)

        # Created response
        rsp = jsonify(episodes)

        # Add link header to response
        rsp.headers.extend(pagination)

        # Add series ID header
        rsp.headers.extend({'Series-ID': show_id})
        return rsp

    @api.response(200, 'Successfully forgotten all episodes from show', model=base_message_schema)
    @api.doc(
        description='Delete all show episodes via its ID. Deleting an episode will mark it as wanted again',
        parser=delete_parser,
    )
    def delete(self, show_id, session):
        """Deletes all episodes of a show"""
        try:
            show = db.show_by_id(show_id, session=session)
        except NoResultFound:
            raise NotFoundError('show with ID %s not found' % show_id)
        args = delete_parser.parse_args()
        forget = args.get('forget')
        for episode in show.episodes:
            db.remove_series_entity(show.name, episode.identifier, forget)
        return success_response('successfully removed all series %s episodes from DB' % show_id)


@api.response(NotFoundError)
@api.response(BadRequest)
@series_api.route('/<int:show_id>/episodes/<int:ep_id>/')
@api.doc(params={'show_id': 'ID of the show', 'ep_id': 'Episode ID'})
class SeriesEpisodeAPI(APIResource):
    @etag
    @api.response(200, 'Episode retrieved successfully for show', episode_schema)
    @api.doc(description='Get a specific episode via its ID and show ID')
    def get(self, show_id, ep_id, session):
        """Get episode by show ID and episode ID"""
        try:
            db.show_by_id(show_id, session=session)
        except NoResultFound:
            raise NotFoundError('show with ID %s not found' % show_id)
        try:
            episode = db.episode_by_id(ep_id, session)
        except NoResultFound:
            raise NotFoundError('episode with ID %s not found' % ep_id)
        if not db.episode_in_show(show_id, ep_id):
            raise BadRequest('episode with id %s does not belong to show %s' % (ep_id, show_id))

        rsp = jsonify(episode.to_dict())

        # Add Series-ID header
        rsp.headers.extend({'Series-ID': show_id})
        return rsp

    @api.response(200, 'Episode successfully forgotten for show', model=base_message_schema)
    @api.doc(
        description='Delete a specific episode via its ID and show ID. Deleting an episode will mark it as '
        'wanted again',
        parser=delete_parser,
    )
    def delete(self, show_id, ep_id, session):
        """Forgets episode by show ID and episode ID"""
        try:
            show = db.show_by_id(show_id, session=session)
        except NoResultFound:
            raise NotFoundError('show with ID %s not found' % show_id)
        try:
            episode = db.episode_by_id(ep_id, session)
        except NoResultFound:
            raise NotFoundError('episode with ID %s not found' % ep_id)
        if not db.episode_in_show(show_id, ep_id):
            raise BadRequest('episode with id %s does not belong to show %s' % (ep_id, show_id))

        args = delete_parser.parse_args()
        db.remove_series_entity(show.name, episode.identifier, args.get('forget'))

        return success_response('successfully removed episode %s from show %s' % (ep_id, show_id))


release_base_parser = api.parser()
release_base_parser.add_argument(
    'downloaded', type=inputs.boolean, help='Filter between release status'
)

sort_choices = ('first_seen', 'downloaded', 'proper_count', 'title')
release_list_parser = api.pagination_parser(release_base_parser, sort_choices)

release_delete_parser = release_base_parser.copy()
release_delete_parser.add_argument(
    'forget',
    type=inputs.boolean,
    default=False,
    help="Enabling this will for 'forget' event that will delete the downloaded"
    " releases from the entire DB, enabling to re-download them",
)


@api.response(NotFoundError)
@api.response(BadRequest)
@series_api.route('/<int:show_id>/seasons/<int:season_id>/releases/')
@api.doc(
    params={'show_id': 'ID of the show', 'season_id': 'Seasons ID'},
    description='Releases are any seen entries that match the seasons. \n'
    'The \'Series-ID\' header will be appended to the result headers.\n'
    'The \'Season-ID\' header will be appended to the result headers.',
)
class SeriesSeasonsReleasesAPI(APIResource):
    @etag
    @api.response(200, 'Releases retrieved successfully for season', season_release_list_schema)
    @api.doc(
        description='Get all matching releases for a specific season of a specific show.',
        parser=release_list_parser,
    )
    def get(self, show_id, season_id, session):
        """Get all season releases by show ID and season ID"""
        try:
            db.show_by_id(show_id, session=session)
        except NoResultFound:
            raise NotFoundError('show with ID %s not found' % show_id)
        try:
            season = db.season_by_id(season_id, session)
        except NoResultFound:
            raise NotFoundError('season with ID %s not found' % season_id)
        if not db.season_in_show(show_id, season_id):
            raise BadRequest(
                'seasons with id %s does not belong to show %s' % (season_id, show_id)
            )

        args = release_list_parser.parse_args()
        # Filter params
        downloaded = args.get('downloaded')

        # Pagination and sorting params
        page = args['page']
        per_page = args['per_page']
        sort_by = args['sort_by']
        sort_order = args['order']

        descending = sort_order == 'desc'

        # Handle max size limit
        if per_page > 100:
            per_page = 100

        start = per_page * (page - 1)
        stop = start + per_page

        kwargs = {
            'downloaded': downloaded,
            'start': start,
            'stop': stop,
            'sort_by': sort_by,
            'descending': descending,
            'season': season,
            'session': session,
        }

        # Total number of releases
        total_items = db.get_season_releases(count=True, **kwargs)

        # Release items
        release_items = [release.to_dict() for release in db.get_season_releases(**kwargs)]

        # Total number of pages
        total_pages = int(ceil(total_items / float(per_page)))

        if total_pages < page and total_pages != 0:
            raise NotFoundError('page %s does not exist' % page)

        # Actual results in page
        actual_size = min(per_page, len(release_items))

        # Get pagination headers
        pagination = pagination_headers(total_pages, total_items, actual_size, request)

        # Created response
        rsp = jsonify(release_items)

        # Add link header to response
        rsp.headers.extend(pagination)

        # Add Series-ID and Episode-ID headers
        rsp.headers.extend({'Series-ID': show_id, 'Season-ID': season_id})

        return rsp

    @api.response(200, 'Successfully deleted all releases for season', model=base_message_schema)
    @api.doc(
        description='Delete all releases for a specific season of a specific show.',
        parser=release_delete_parser,
    )
    def delete(self, show_id, season_id, session):
        """Deletes all season releases by show ID and season ID"""
        try:
            db.show_by_id(show_id, session=session)
        except NoResultFound:
            raise NotFoundError('show with ID %s not found' % show_id)
        try:
            season = db.season_by_id(season_id, session)
        except NoResultFound:
            raise NotFoundError('seasons with ID %s not found' % season_id)
        if not db.season_in_show(show_id, season_id):
            raise BadRequest('season with id %s does not belong to show %s' % (season_id, show_id))

        args = release_delete_parser.parse_args()
        downloaded = args.get('downloaded') is True if args.get('downloaded') is not None else None
        release_items = []
        for release in season.releases:
            if (
                downloaded
                and release.downloaded
                or downloaded is False
                and not release.downloaded
                or not downloaded
            ):
                release_items.append(release)

        for release in release_items:
            if args.get('forget'):
                fire_event('forget', release.title)
            db.delete_season_release_by_id(release.id)
        return success_response(
            'successfully deleted all releases for season %s from show %s' % (season_id, show_id)
        )

    @api.response(
        200, 'Successfully reset all downloaded releases for season', model=base_message_schema
    )
    @api.doc(
        description='Resets all of the downloaded releases of an season, clearing the quality to be downloaded '
        'again,'
    )
    def put(self, show_id, season_id, session):
        """Marks all downloaded season releases as not downloaded"""
        try:
            db.show_by_id(show_id, session=session)
        except NoResultFound:
            raise NotFoundError('show with ID %s not found' % show_id)
        try:
            season = db.season_by_id(season_id, session)
        except NoResultFound:
            raise NotFoundError('season with ID %s not found' % season_id)
        if not db.season_in_show(show_id, season_id):
            raise BadRequest('season with id %s does not belong to show %s' % (season_id, show_id))

        for release in season.releases:
            if release.downloaded:
                release.downloaded = False

        return success_response(
            'successfully reset download status for all releases for season %s from show %s'
            % (season_id, show_id)
        )


@api.response(NotFoundError)
@api.response(BadRequest)
@series_api.route('/<int:show_id>/seasons/<int:season_id>/releases/<int:rel_id>/')
@api.doc(
    params={'show_id': 'ID of the show', 'season_id': 'Season ID', 'rel_id': 'Release ID'},
    description='Releases are any seen entries that match the season. \n'
    'The \'Series-ID\' header will be appended to the result headers.\n'
    'The \'Season-ID\' header will be appended to the result headers.',
)
class SeriesSeasonReleaseAPI(APIResource):
    @etag
    @api.response(200, 'Release retrieved successfully for season', season_release_schema)
    @api.doc(
        description='Get a specific downloaded release for a specific season of a specific show'
    )
    def get(self, show_id, season_id, rel_id, session):
        """Get season release by show ID, season ID and release ID"""
        try:
            db.show_by_id(show_id, session=session)
        except NoResultFound:
            raise NotFoundError('show with ID %s not found' % show_id)
        try:
            db.season_by_id(season_id, session)
        except NoResultFound:
            raise NotFoundError('season with ID %s not found' % season_id)
        try:
            release = db.season_release_by_id(rel_id, session)
        except NoResultFound:
            raise NotFoundError('release with ID %s not found' % rel_id)
        if not db.season_in_show(show_id, season_id):
            raise BadRequest('season with id %s does not belong to show %s' % (season_id, show_id))
        if not db.release_in_season(season_id, rel_id):
            raise BadRequest('release id %s does not belong to season %s' % (rel_id, season_id))

        rsp = jsonify(release.to_dict())
        rsp.headers.extend({'Series-ID': show_id, 'Season-ID': season_id})
        return rsp

    @api.response(200, 'Release successfully deleted', model=base_message_schema)
    @api.doc(
        description='Delete a specific releases for a specific season of a specific show.',
        parser=delete_parser,
    )
    def delete(self, show_id, season_id, rel_id, session):
        """Delete episode release by show ID, season ID and release ID"""
        try:
            db.show_by_id(show_id, session=session)
        except NoResultFound:
            raise NotFoundError('show with ID %s not found' % show_id)
        try:
            db.season_by_id(season_id, session)
        except NoResultFound:
            raise NotFoundError('season with ID %s not found' % season_id)
        try:
            release = db.season_release_by_id(rel_id, session)
        except NoResultFound:
            raise NotFoundError('release with ID %s not found' % rel_id)
        if not db.season_in_show(show_id, season_id):
            raise BadRequest('season with id %s does not belong to show %s' % (season_id, show_id))
        if not db.release_in_season(season_id, rel_id):
            raise BadRequest('release id %s does not belong to season %s' % (rel_id, season_id))
        args = delete_parser.parse_args()
        if args.get('forget'):
            fire_event('forget', release.title)

        db.delete_season_release_by_id(rel_id)
        return success_response(
            'successfully deleted release %d from season %d' % (rel_id, season_id)
        )

    @api.response(200, 'Successfully reset downloaded release status', model=season_release_schema)
    @api.doc(
        description='Resets the downloaded release status, clearing the quality to be downloaded again'
    )
    def put(self, show_id, season_id, rel_id, session):
        """Resets a downloaded release status"""
        try:
            db.show_by_id(show_id, session=session)
        except NoResultFound:
            raise NotFoundError('show with ID %s not found' % show_id)
        try:
            db.season_by_id(season_id, session)
        except NoResultFound:
            raise NotFoundError('season with ID %s not found' % season_id)
        try:
            release = db.season_release_by_id(rel_id, session)
        except NoResultFound:
            raise NotFoundError('release with ID %s not found' % rel_id)
        if not db.season_in_show(show_id, season_id):
            raise BadRequest('season with id %s does not belong to show %s' % (season_id, show_id))
        if not db.release_in_season(season_id, rel_id):
            raise BadRequest('release id %s does not belong to episode %s' % (rel_id, season_id))

        if not release.downloaded:
            raise BadRequest('release with id %s is not set as downloaded' % rel_id)
        release.downloaded = False

        rsp = jsonify(release.to_dict())
        rsp.headers.extend({'Series-ID': show_id, 'Season-ID': season_id})
        return rsp


@api.response(NotFoundError)
@api.response(BadRequest)
@series_api.route('/<int:show_id>/episodes/<int:ep_id>/releases/')
@api.doc(
    params={'show_id': 'ID of the show', 'ep_id': 'Episode ID'},
    description='Releases are any seen entries that match the episode. \n'
    'The \'Series-ID\' header will be appended to the result headers.\n'
    'The \'Episode-ID\' header will be appended to the result headers.',
)
class SeriesEpisodeReleasesAPI(APIResource):
    @etag
    @api.response(200, 'Releases retrieved successfully for episode', episode_release_list_schema)
    @api.doc(
        description='Get all matching releases for a specific episode of a specific show.',
        parser=release_list_parser,
    )
    def get(self, show_id, ep_id, session):
        """Get all episodes releases by show ID and episode ID"""
        try:
            db.show_by_id(show_id, session=session)
        except NoResultFound:
            raise NotFoundError('show with ID %s not found' % show_id)
        try:
            episode = db.episode_by_id(ep_id, session)
        except NoResultFound:
            raise NotFoundError('episode with ID %s not found' % ep_id)
        if not db.episode_in_show(show_id, ep_id):
            raise BadRequest('episode with id %s does not belong to show %s' % (ep_id, show_id))

        args = release_list_parser.parse_args()
        # Filter params
        downloaded = args.get('downloaded')

        # Pagination and sorting params
        page = args['page']
        per_page = args['per_page']
        sort_by = args['sort_by']
        sort_order = args['order']

        descending = sort_order == 'desc'

        # Handle max size limit
        if per_page > 100:
            per_page = 100

        start = per_page * (page - 1)
        stop = start + per_page

        kwargs = {
            'downloaded': downloaded,
            'start': start,
            'stop': stop,
            'sort_by': sort_by,
            'descending': descending,
            'episode': episode,
            'session': session,
        }

        # Total number of releases
        total_items = db.get_episode_releases(count=True, **kwargs)

        # Release items
        release_items = [release.to_dict() for release in db.get_episode_releases(**kwargs)]

        # Total number of pages
        total_pages = int(ceil(total_items / float(per_page)))

        if total_pages < page and total_pages != 0:
            raise NotFoundError('page %s does not exist' % page)

        # Actual results in page
        actual_size = min(per_page, len(release_items))

        # Get pagination headers
        pagination = pagination_headers(total_pages, total_items, actual_size, request)

        # Created response
        rsp = jsonify(release_items)

        # Add link header to response
        rsp.headers.extend(pagination)

        # Add Series-ID and Episode-ID headers
        rsp.headers.extend({'Series-ID': show_id, 'Episode-ID': ep_id})

        return rsp

    @api.response(200, 'Successfully deleted all releases for episode', model=base_message_schema)
    @api.doc(
        description='Delete all releases for a specific episode of a specific show.',
        parser=release_delete_parser,
    )
    def delete(self, show_id, ep_id, session):
        """Deletes all episodes releases by show ID and episode ID"""
        try:
            db.show_by_id(show_id, session=session)
        except NoResultFound:
            raise NotFoundError('show with ID %s not found' % show_id)
        try:
            episode = db.episode_by_id(ep_id, session)
        except NoResultFound:
            raise NotFoundError('episode with ID %s not found' % ep_id)
        if not db.episode_in_show(show_id, ep_id):
            raise BadRequest('episode with id %s does not belong to show %s' % (ep_id, show_id))

        args = release_delete_parser.parse_args()
        downloaded = args.get('downloaded') is True if args.get('downloaded') is not None else None
        release_items = []
        for release in episode.releases:
            if (
                downloaded
                and release.downloaded
                or downloaded is False
                and not release.downloaded
                or not downloaded
            ):
                release_items.append(release)

        for release in release_items:
            if args.get('forget'):
                fire_event('forget', release.title)
            db.delete_episode_release_by_id(release.id)
        return success_response(
            'successfully deleted all releases for episode %s from show %s' % (ep_id, show_id)
        )

    @api.response(
        200, 'Successfully reset all downloaded releases for episode', model=base_message_schema
    )
    @api.doc(
        description='Resets all of the downloaded releases of an episode, clearing the quality to be downloaded '
        'again,'
    )
    def put(self, show_id, ep_id, session):
        """Marks all downloaded releases as not downloaded"""
        try:
            db.show_by_id(show_id, session=session)
        except NoResultFound:
            raise NotFoundError('show with ID %s not found' % show_id)
        try:
            episode = db.episode_by_id(ep_id, session)
        except NoResultFound:
            raise NotFoundError('episode with ID %s not found' % ep_id)
        if not db.episode_in_show(show_id, ep_id):
            raise BadRequest('episode with id %s does not belong to show %s' % (ep_id, show_id))

        for release in episode.releases:
            if release.downloaded:
                release.downloaded = False

        return success_response(
            'successfully reset download status for all releases for episode %s from show %s'
            % (ep_id, show_id)
        )


@api.response(NotFoundError)
@api.response(BadRequest)
@series_api.route('/<int:show_id>/episodes/<int:ep_id>/releases/<int:rel_id>/')
@api.doc(
    params={'show_id': 'ID of the show', 'ep_id': 'Episode ID', 'rel_id': 'Release ID'},
    description='Releases are any seen entries that match the episode. \n'
    'The \'Series-ID\' header will be appended to the result headers.\n'
    'The \'Episode-ID\' header will be appended to the result headers.',
)
class SeriesEpisodeReleaseAPI(APIResource):
    @etag
    @api.response(200, 'Release retrieved successfully for episode', episode_release_schema)
    @api.doc(
        description='Get a specific downloaded release for a specific episode of a specific show'
    )
    def get(self, show_id, ep_id, rel_id, session):
        """Get episode release by show ID, episode ID and release ID"""
        try:
            db.show_by_id(show_id, session=session)
        except NoResultFound:
            raise NotFoundError('show with ID %s not found' % show_id)
        try:
            db.episode_by_id(ep_id, session)
        except NoResultFound:
            raise NotFoundError('episode with ID %s not found' % ep_id)
        try:
            release = db.episode_release_by_id(rel_id, session)
        except NoResultFound:
            raise NotFoundError('release with ID %s not found' % rel_id)
        if not db.episode_in_show(show_id, ep_id):
            raise BadRequest('episode with id %s does not belong to show %s' % (ep_id, show_id))
        if not db.release_in_episode(ep_id, rel_id):
            raise BadRequest('release id %s does not belong to episode %s' % (rel_id, ep_id))

        rsp = jsonify(release.to_dict())
        rsp.headers.extend({'Series-ID': show_id, 'Episode-ID': ep_id})
        return rsp

    @api.response(200, 'Release successfully deleted', model=base_message_schema)
    @api.doc(
        description='Delete a specific releases for a specific episode of a specific show.',
        parser=delete_parser,
    )
    def delete(self, show_id, ep_id, rel_id, session):
        """Delete episode release by show ID, episode ID and release ID"""
        try:
            db.show_by_id(show_id, session=session)
        except NoResultFound:
            raise NotFoundError('show with ID %s not found' % show_id)
        try:
            db.episode_by_id(ep_id, session)
        except NoResultFound:
            raise NotFoundError('episode with ID %s not found' % ep_id)
        try:
            release = db.episode_release_by_id(rel_id, session)
        except NoResultFound:
            raise NotFoundError('release with ID %s not found' % rel_id)
        if not db.episode_in_show(show_id, ep_id):
            raise BadRequest('episode with id %s does not belong to show %s' % (ep_id, show_id))
        if not db.release_in_episode(ep_id, rel_id):
            raise BadRequest('release id %s does not belong to episode %s' % (rel_id, ep_id))
        args = delete_parser.parse_args()
        if args.get('forget'):
            fire_event('forget', release.title)

        db.delete_episode_release_by_id(rel_id)
        return success_response(
            'successfully deleted release %d from episode %d' % (rel_id, ep_id)
        )

    @api.response(
        200, 'Successfully reset downloaded release status', model=episode_release_schema
    )
    @api.doc(
        description='Resets the downloaded release status, clearing the quality to be downloaded again'
    )
    def put(self, show_id, ep_id, rel_id, session):
        """Resets a downloaded release status"""
        try:
            db.show_by_id(show_id, session=session)
        except NoResultFound:
            raise NotFoundError('show with ID %s not found' % show_id)
        try:
            db.episode_by_id(ep_id, session)
        except NoResultFound:
            raise NotFoundError('episode with ID %s not found' % ep_id)
        try:
            release = db.episode_release_by_id(rel_id, session)
        except NoResultFound:
            raise NotFoundError('release with ID %s not found' % rel_id)
        if not db.episode_in_show(show_id, ep_id):
            raise BadRequest('episode with id %s does not belong to show %s' % (ep_id, show_id))
        if not db.release_in_episode(ep_id, rel_id):
            raise BadRequest('release id %s does not belong to episode %s' % (rel_id, ep_id))

        if not release.downloaded:
            raise BadRequest('release with id %s is not set as downloaded' % rel_id)
        release.downloaded = False

        rsp = jsonify(release.to_dict())
        rsp.headers.extend({'Series-ID': show_id, 'Episode-ID': ep_id})
        return rsp
